# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import codecs
import itertools
import json
from optparse import OptionParser
import os
import sqlite3
import sys
from collections import defaultdict

import urllib

from swift.common.exceptions import LockTimeout
from swift.common.utils import hash_path, storage_directory, \
    Timestamp, is_valid_ipv6
from swift.common.ring import Ring
from swift.common.request_helpers import is_sys_meta, is_user_meta, \
    strip_sys_meta_prefix, strip_user_meta_prefix, \
    is_object_transient_sysmeta, strip_object_transient_sysmeta_prefix
from swift.account.backend import AccountBroker, DATADIR as ABDATADIR
from swift.container.backend import ContainerBroker, DATADIR as CBDATADIR
from swift.obj.diskfile import get_data_dir, read_metadata, DATADIR_BASE, \
    extract_policy
from swift.common.storage_policy import POLICIES, reload_storage_policies
from swift.common.swob import wsgi_to_str
from swift.common.middleware.crypto.crypto_utils import load_crypto_meta
from swift.common.utils import md5, set_swift_dir


class InfoSystemExit(Exception):
    """
    Indicates to the caller that a sys.exit(1) should be performed.
    """
    pass


def parse_get_node_args(options, args):
    """
    Parse the get_nodes commandline args

    :returns: a tuple, (ring_path, args)
    """
    ring_path = None

    if options.policy_name:
        if POLICIES.get_by_name(options.policy_name) is None:
            raise InfoSystemExit('No policy named %r' % options.policy_name)
    elif args and args[0].endswith('.ring.gz'):
        if os.path.exists(args[0]):
            ring_path = args.pop(0)
        else:
            raise InfoSystemExit('Ring file does not exist')

    if options.quoted:
        args = [urllib.parse.unquote(arg) for arg in args]
    if len(args) == 1:
        args = args[0].strip('/').split('/', 2)

    if not ring_path and not options.policy_name:
        raise InfoSystemExit('Need to specify policy_name or <ring.gz>')

    if not (args or options.partition):
        raise InfoSystemExit('No target specified')

    if len(args) > 3:
        raise InfoSystemExit('Invalid arguments')

    return ring_path, args


def curl_head_command(ip, port, device, part, target, policy_index):
    """
    Provide a string that is a well formatted curl command to HEAD an object
    on a storage node.

    :param ip: the ip of the node
    :param port: the port of the node
    :param device: the device of the node
    :param target: the path of the target resource
    :param policy_index: the policy_index of the target resource (can be None)

    :returns: a string, a well formatted curl command
    """
    if is_valid_ipv6(ip):
        formatted_ip = '[%s]' % ip
    else:
        formatted_ip = ip

    cmd = 'curl -g -I -XHEAD "http://%s:%s/%s/%s/%s"' % (
        formatted_ip, port, device, part, urllib.parse.quote(target))
    if policy_index is not None:
        cmd += ' -H "%s: %s"' % ('X-Backend-Storage-Policy-Index',
                                 policy_index)
    cmd += ' --path-as-is'
    return cmd


def print_ring_locations(ring, datadir, account, container=None, obj=None,
                         tpart=None, all_nodes=False, policy_index=None):
    """
    print out ring locations of specified type

    :param ring: ring instance
    :param datadir: name of directory where things are stored. Usually one of
                    "accounts", "containers", "objects", or "objects-N".
    :param account: account name
    :param container: container name
    :param obj: object name
    :param tpart: target partition in ring
    :param all_nodes: include all handoff nodes. If false, only the N primary
                      nodes and first N handoffs will be printed.
    :param policy_index: include policy_index in curl headers
    """
    if not ring:
        raise ValueError("No ring specified")
    if not datadir:
        raise ValueError("No datadir specified")
    if tpart is None and not account:
        raise ValueError("No partition or account/container/object specified")
    if not account and (container or obj):
        raise ValueError("Container/object specified without account")
    if obj and not container:
        raise ValueError('Object specified without container')

    if obj:
        target = '%s/%s/%s' % (account, container, obj)
    elif container:
        target = '%s/%s' % (account, container)
    else:
        target = '%s' % (account)

    if tpart:
        part = int(tpart)
    else:
        part = ring.get_part(account, container, obj)

    primary_nodes = ring.get_part_nodes(part)
    handoff_nodes = ring.get_more_nodes(part)
    if not all_nodes:
        handoff_nodes = itertools.islice(handoff_nodes, len(primary_nodes))
    handoff_nodes = list(handoff_nodes)

    if account and not tpart:
        path_hash = hash_path(account, container, obj)
    else:
        path_hash = None
    print('Partition\t%s' % part)
    print('Hash     \t%s\n' % path_hash)

    for node in primary_nodes:
        print('Server:Port Device\t%s:%s %s' % (node['ip'], node['port'],
                                                node['device']))
    for node in handoff_nodes:
        print('Server:Port Device\t%s:%s %s\t [Handoff]' % (
            node['ip'], node['port'], node['device']))

    print("\n")

    for node in primary_nodes:
        cmd = curl_head_command(node['ip'], node['port'], node['device'],
                                part, target, policy_index)
        print(cmd)
    for node in handoff_nodes:
        cmd = curl_head_command(node['ip'], node['port'], node['device'],
                                part, target, policy_index)
        cmd += ' # [Handoff]'
        print(cmd)

    print("\n\nUse your own device location of servers:")
    print("such as \"export DEVICE=/srv/node\"")
    if path_hash:
        for node in primary_nodes:
            print('ssh %s "ls -lah ${DEVICE:-/srv/node*}/%s/%s"' %
                  (node['ip'], node['device'],
                   storage_directory(datadir, part, path_hash)))
        for node in handoff_nodes:
            print('ssh %s "ls -lah ${DEVICE:-/srv/node*}/%s/%s" # [Handoff]' %
                  (node['ip'], node['device'],
                   storage_directory(datadir, part, path_hash)))
    else:
        for node in primary_nodes:
            print('ssh %s "ls -lah ${DEVICE:-/srv/node*}/%s/%s/%d"' %
                  (node['ip'], node['device'], datadir, part))
        for node in handoff_nodes:
            print('ssh %s "ls -lah ${DEVICE:-/srv/node*}/%s/%s/%d"'
                  ' # [Handoff]' %
                  (node['ip'], node['device'], datadir, part))

    print('\nnote: `/srv/node*` is used as default value of `devices`, the '
          'real value is set in the config file on each storage node.')


def get_max_len_sync_item(syncs, item, title):
    def map_func(element):
        return str(element[item])
    return max(list(map(len, map(map_func, syncs))) + [len(title)])


def print_db_syncs(incoming, syncs):
    max_sync_point_len = get_max_len_sync_item(syncs, 'sync_point',
                                               "Sync Point")
    max_remote_len = get_max_len_sync_item(syncs, 'remote_id', "Remote ID")
    print('%s Syncs:' % ('Incoming' if incoming else 'Outgoing'))
    print('  %s\t%s\t%s' % ("Sync Point".ljust(max_sync_point_len),
                            "Remote ID".ljust(max_remote_len),
                            "Updated At"))
    for sync in syncs:
        print('  %s\t%s\t%s (%s)' % (
            str(sync['sync_point']).ljust(max_sync_point_len),
            sync['remote_id'].ljust(max_remote_len),
            Timestamp(sync['updated_at']).isoformat,
            sync['updated_at']))


def print_db_info_metadata(db_type, info, metadata, drop_prefixes=False,
                           verbose=False):
    """
    print out data base info/metadata based on its type

    :param db_type: database type, account or container
    :param info: dict of data base info
    :param metadata: dict of data base metadata
    :param drop_prefixes: if True, strip "X-Account-Meta-",
                          "X-Container-Meta-", "X-Account-Sysmeta-", and
                          "X-Container-Sysmeta-" when displaying
                          User Metadata and System Metadata dicts
    """
    if info is None:
        raise ValueError('DB info is None')

    if db_type not in ['container', 'account']:
        raise ValueError('Wrong DB type')

    try:
        account = info['account']
        container = None

        if db_type == 'container':
            container = info['container']
            path = '/%s/%s' % (account, container)
        else:
            path = '/%s' % account

        print('Path: %s' % path)
        print('  Account: %s' % account)

        if db_type == 'container':
            print('  Container: %s' % container)

        print('  Deleted: %s' % info['is_deleted'])
        path_hash = hash_path(account, container)
        if db_type == 'container':
            print('  Container Hash: %s' % path_hash)
        else:
            print('  Account Hash: %s' % path_hash)

        print('Metadata:')
        print('  Created at: %s (%s)' %
              (Timestamp(info['created_at']).isoformat,
               info['created_at']))
        print('  Put Timestamp: %s (%s)' %
              (Timestamp(info['put_timestamp']).isoformat,
               info['put_timestamp']))
        print('  Delete Timestamp: %s (%s)' %
              (Timestamp(info['delete_timestamp']).isoformat,
               info['delete_timestamp']))
        print('  Status Timestamp: %s (%s)' %
              (Timestamp(info['status_changed_at']).isoformat,
               info['status_changed_at']))
        if db_type == 'account':
            print('  Container Count: %s' % info['container_count'])
        print('  Object Count: %s' % info['object_count'])
        print('  Bytes Used: %s' % info['bytes_used'])
        if db_type == 'container':
            try:
                policy_name = POLICIES[info['storage_policy_index']].name
            except KeyError:
                policy_name = 'Unknown'
            print('  Storage Policy: %s (%s)' % (
                policy_name, info['storage_policy_index']))
            print('  Reported Put Timestamp: %s (%s)' %
                  (Timestamp(info['reported_put_timestamp']).isoformat,
                   info['reported_put_timestamp']))
            print('  Reported Delete Timestamp: %s (%s)' %
                  (Timestamp(info['reported_delete_timestamp']).isoformat,
                   info['reported_delete_timestamp']))
            print('  Reported Object Count: %s' %
                  info['reported_object_count'])
            print('  Reported Bytes Used: %s' % info['reported_bytes_used'])
        print('  Chexor: %s' % info['hash'])
        print('  UUID: %s' % info['id'])
    except KeyError as e:
        raise ValueError('Info is incomplete: %s' % e)

    meta_prefix = 'x_' + db_type + '_'
    for key, value in info.items():
        if key.lower().startswith(meta_prefix):
            title = key.replace('_', '-').title()
            print('  %s: %s' % (title, value))
    user_metadata = {}
    sys_metadata = {}
    for key, (value, timestamp) in metadata.items():
        if is_user_meta(db_type, key):
            if drop_prefixes:
                key = strip_user_meta_prefix(db_type, key)
            user_metadata[key] = value
        elif is_sys_meta(db_type, key):
            if drop_prefixes:
                key = strip_sys_meta_prefix(db_type, key)
            sys_metadata[key] = value
        else:
            title = key.replace('_', '-').title()
            print('  %s: %s' % (title, value))
    if sys_metadata:
        print('  System Metadata:')
        for key, value in sys_metadata.items():
            print('    %s: %s' % (key, value))
    else:
        print('No system metadata found in db file')

    if user_metadata:
        print('  User Metadata:')
        for key, value in user_metadata.items():
            print('    %s: %s' % (key, value))
    else:
        print('No user metadata found in db file')

    if db_type == 'container':
        print('Sharding Metadata:')
        shard_type = 'root' if info['is_root'] else 'shard'
        print('  Type: %s' % shard_type)
        print('  State: %s' % info['db_state'])
    if info.get('shard_ranges'):
        num_shards = len(info['shard_ranges'])
        print('Shard Ranges (%d):' % num_shards)
        count_by_state = defaultdict(int)
        for srange in info['shard_ranges']:
            count_by_state[(srange.state, srange.state_text)] += 1
        print('  States:')
        for key_state, count in sorted(count_by_state.items()):
            key, state = key_state
            print('    %9s: %s' % (state, count))
        if verbose:
            for srange in info['shard_ranges']:
                srange = dict(srange, state_text=srange.state_text)
                print('  Name: %(name)s' % srange)
                print('    lower: %(lower)r, upper: %(upper)r' % srange)
                print('    Object Count: %(object_count)d, Bytes Used: '
                      '%(bytes_used)d, State: %(state_text)s (%(state)d)'
                      % srange)
                print('    Created at: %s (%s)'
                      % (Timestamp(srange['timestamp']).isoformat,
                         srange['timestamp']))
                print('    Meta Timestamp: %s (%s)'
                      % (Timestamp(srange['meta_timestamp']).isoformat,
                         srange['meta_timestamp']))
        else:
            print('(Use -v/--verbose to show more Shard Ranges details)')


def print_obj_metadata(metadata, drop_prefixes=False):
    """
    Print out basic info and metadata from object, as returned from
    :func:`swift.obj.diskfile.read_metadata`.

    Metadata should include the keys: name, Content-Type, and
    X-Timestamp.

    Additional metadata is displayed unmodified.

    :param metadata: dict of object metadata
    :param drop_prefixes: if True, strip "X-Object-Meta-", "X-Object-Sysmeta-",
                          and "X-Object-Transient-Sysmeta-" when displaying
                          User Metadata, System Metadata, and Transient
                          System Metadata entries

    :raises ValueError:
    """
    user_metadata = {}
    sys_metadata = {}
    transient_sys_metadata = {}
    other_metadata = {}

    if not metadata:
        raise ValueError('Metadata is None')
    path = metadata.pop('name', '')
    content_type = metadata.pop('Content-Type', '')
    ts = Timestamp(metadata.pop('X-Timestamp', Timestamp.zero()))
    account = container = obj = obj_hash = None
    if path:
        try:
            account, container, obj = path.split('/', 3)[1:]
        except ValueError:
            raise ValueError('Path is invalid for object %r' % path)
        else:
            obj_hash = hash_path(account, container, obj)
        print('Path: %s' % path)
        print('  Account: %s' % account)
        print('  Container: %s' % container)
        print('  Object: %s' % obj)
        print('  Object hash: %s' % obj_hash)
    else:
        print('Path: Not found in metadata')
    if content_type:
        print('Content-Type: %s' % content_type)
    else:
        print('Content-Type: Not found in metadata')
    if ts:
        print('Timestamp: %s (%s)' % (ts.isoformat, ts.internal))
    else:
        print('Timestamp: Not found in metadata')

    for key, value in metadata.items():
        if is_user_meta('Object', key):
            if drop_prefixes:
                key = strip_user_meta_prefix('Object', key)
            user_metadata[key] = value
        elif is_sys_meta('Object', key):
            if drop_prefixes:
                key = strip_sys_meta_prefix('Object', key)
            sys_metadata[key] = value
        elif is_object_transient_sysmeta(key):
            if drop_prefixes:
                key = strip_object_transient_sysmeta_prefix(key)
            transient_sys_metadata[key] = value
        else:
            other_metadata[key] = value

    def print_metadata(title, items):
        print(title)
        if items:
            for key, value in sorted(items.items()):
                print('  %s: %s' % (key, value))
        else:
            print('  No metadata found')

    print_metadata('System Metadata:', sys_metadata)
    print_metadata('Transient System Metadata:', transient_sys_metadata)
    print_metadata('User Metadata:', user_metadata)
    print_metadata('Other Metadata:', other_metadata)
    for label, meta in [
        ('Data crypto details',
         sys_metadata.get('X-Object-Sysmeta-Crypto-Body-Meta')),
        ('Metadata crypto details',
         transient_sys_metadata.get('X-Object-Transient-Sysmeta-Crypto-Meta')),
    ]:
        if meta is None:
            continue
        print('%s: %s' % (
            label,
            json.dumps(load_crypto_meta(meta, b64decode=False), indent=2,
                       sort_keys=True, separators=(',', ': '))))


def print_info(db_type, db_file, swift_dir='/etc/swift', stale_reads_ok=False,
               drop_prefixes=False, verbose=False, sync=False):
    if db_type not in ('account', 'container'):
        print("Unrecognized DB type: internal error")
        raise InfoSystemExit()
    if not os.path.exists(db_file) or not db_file.endswith('.db'):
        print("DB file doesn't exist")
        raise InfoSystemExit()
    if not db_file.startswith(('/', './')):
        db_file = './' + db_file  # don't break if the bare db file is given
    if db_type == 'account':
        broker = AccountBroker(db_file, stale_reads_ok=stale_reads_ok)
        datadir = ABDATADIR
    else:
        broker = ContainerBroker(db_file, stale_reads_ok=stale_reads_ok)
        datadir = CBDATADIR
    try:
        info = broker.get_info()
    except sqlite3.OperationalError as err:
        if 'no such table' in str(err):
            print("Does not appear to be a DB of type \"%s\": %s"
                  % (db_type, db_file))
            raise InfoSystemExit()
        raise
    account = info['account']
    container = None
    info['is_deleted'] = broker.is_deleted()
    if db_type == 'container':
        container = info['container']
        info['is_root'] = broker.is_root_container()
        sranges = broker.get_shard_ranges()
        if sranges:
            info['shard_ranges'] = sranges
    print_db_info_metadata(
        db_type, info, broker.metadata, drop_prefixes, verbose)
    if sync:
        # Print incoming / outgoing sync tables.
        for incoming in (True, False):
            print_db_syncs(incoming, broker.get_syncs(incoming,
                                                      include_timestamp=True))
    try:
        ring = Ring(swift_dir, ring_name=db_type)
    except Exception:
        ring = None
    else:
        print_ring_locations(ring, datadir, account, container)


def print_obj(datafile, check_etag=True, swift_dir='/etc/swift',
              policy_name='', drop_prefixes=False):
    """
    Display information about an object read from the datafile.
    Optionally verify the datafile content matches the ETag metadata.

    :param datafile: path on disk to object file
    :param check_etag: boolean, will read datafile content and verify
                       computed checksum matches value stored in
                       metadata.
    :param swift_dir: the path on disk to rings
    :param policy_name: optionally the name to use when finding the ring
    :param drop_prefixes: if True, strip "X-Object-Meta-", "X-Object-Sysmeta-",
                          and "X-Object-Transient-Sysmeta-" when displaying
                          User Metadata, System Metadata, and Transient
                          System Metadata entries
    """
    if not os.path.exists(datafile):
        print("Data file doesn't exist")
        raise InfoSystemExit()
    if not datafile.startswith(('/', './')):
        datafile = './' + datafile

    policy_index = None
    ring = None
    datadir = DATADIR_BASE

    # try to extract policy index from datafile disk path
    fullpath = os.path.abspath(datafile)
    policy_index = int(extract_policy(fullpath) or POLICIES.legacy)

    try:
        if policy_index:
            datadir += '-' + str(policy_index)
            ring = Ring(swift_dir, ring_name='object-' + str(policy_index))
        elif policy_index == 0:
            ring = Ring(swift_dir, ring_name='object')
    except IOError:
        # no such ring
        pass

    if policy_name:
        policy = POLICIES.get_by_name(policy_name)
        if policy:
            policy_index_for_name = policy.idx
            if (policy_index is not None and
               policy_index_for_name is not None and
               policy_index != policy_index_for_name):
                print('WARNING: Ring does not match policy!')
                print('Double check your policy name!')
            if not ring and policy_index_for_name:
                ring = POLICIES.get_object_ring(policy_index_for_name,
                                                swift_dir)
                datadir = get_data_dir(policy_index_for_name)

    with open(datafile, 'rb') as fp:
        try:
            metadata = read_metadata(fp)
        except EOFError:
            print("Invalid metadata")
            raise InfoSystemExit()
        metadata = {wsgi_to_str(k): v if k == 'name' else wsgi_to_str(v)
                    for k, v in metadata.items()}

        etag = metadata.pop('ETag', '')
        length = metadata.pop('Content-Length', '')
        path = metadata.get('name', '')
        print_obj_metadata(metadata, drop_prefixes)

        # Optional integrity check; it's useful, but slow.
        file_len = None
        if check_etag:
            h = md5(usedforsecurity=False)
            file_len = 0
            while True:
                data = fp.read(64 * 1024)
                if not data:
                    break
                h.update(data)
                file_len += len(data)
            h = h.hexdigest()
            if etag:
                if h == etag:
                    print('ETag: %s (valid)' % etag)
                else:
                    print("ETag: %s doesn't match file hash of %s!" %
                          (etag, h))
            else:
                print('ETag: Not found in metadata')
        else:
            print('ETag: %s (not checked)' % etag)
            file_len = os.fstat(fp.fileno()).st_size

        if length:
            if file_len == int(length):
                print('Content-Length: %s (valid)' % length)
            else:
                print("Content-Length: %s doesn't match file length of %s"
                      % (length, file_len))
        else:
            print('Content-Length: Not found in metadata')

        account, container, obj = path.split('/', 3)[1:]
        if ring:
            print_ring_locations(ring, datadir, account, container, obj,
                                 policy_index=policy_index)


def print_item_locations(ring, ring_name=None, account=None, container=None,
                         obj=None, **kwargs):
    """
    Display placement information for an item based on ring lookup.

    If a ring is provided it always takes precedence, but warnings will be
    emitted if it doesn't match other optional arguments like the policy_name
    or ring_name.

    If no ring is provided the ring_name and/or policy_name will be used to
    lookup the ring.

    :param ring: a ring instance
    :param ring_name: server type, or storage policy ring name if object ring
    :param account: account name
    :param container: container name
    :param obj: object name
    :param partition: part number for non path lookups
    :param policy_name: name of storage policy to use to lookup the ring
    :param all_nodes: include all handoff nodes. If false, only the N primary
                      nodes and first N handoffs will be printed.
    """

    policy_name = kwargs.get('policy_name', None)
    part = kwargs.get('partition', None)
    all_nodes = kwargs.get('all', False)
    swift_dir = kwargs.get('swift_dir', '/etc/swift')

    if ring and policy_name:
        policy = POLICIES.get_by_name(policy_name)
        if policy:
            if ring_name != policy.ring_name:
                print('WARNING: mismatch between ring and policy name!')
        else:
            print('WARNING: Policy %s is not valid' % policy_name)

    policy_index = None
    if ring is None and (obj or part):
        if not policy_name:
            print('Need a ring or policy')
            raise InfoSystemExit()
        policy = POLICIES.get_by_name(policy_name)
        if not policy:
            print('No policy named %r' % policy_name)
            raise InfoSystemExit()
        policy_index = int(policy)
        ring = POLICIES.get_object_ring(policy_index, swift_dir)
        ring_name = (POLICIES.get_by_name(policy_name)).ring_name

    if (container or obj) and not account:
        print('No account specified')
        raise InfoSystemExit()

    if obj and not container:
        print('No container specified')
        raise InfoSystemExit()

    if not account and not part:
        print('No target specified')
        raise InfoSystemExit()

    loc = '<type>'
    if part and ring_name:
        if '-' in ring_name and ring_name.startswith('object'):
            loc = 'objects-' + ring_name.split('-', 1)[1]
        else:
            loc = ring_name + 's'
    if account and container and obj:
        loc = 'objects'
        if '-' in ring_name and ring_name.startswith('object'):
            policy_index = int(ring_name.rsplit('-', 1)[1])
            loc = 'objects-%d' % policy_index
    if account and container and not obj:
        loc = 'containers'
        if not any([ring, ring_name]):
            ring = Ring(swift_dir, ring_name='container')
        else:
            if ring_name != 'container':
                print('WARNING: account/container specified ' +
                      'but ring not named "container"')
    if account and not container and not obj:
        loc = 'accounts'
        if not any([ring, ring_name]):
            ring = Ring(swift_dir, ring_name='account')
        else:
            if ring_name != 'account':
                print('WARNING: account specified ' +
                      'but ring not named "account"')

    if account:
        print('\nAccount  \t%s' % urllib.parse.quote(account))
    if container:
        print('Container\t%s' % urllib.parse.quote(container))
    if obj:
        print('Object   \t%s\n\n' % urllib.parse.quote(obj))
    print_ring_locations(ring, loc, account, container, obj, part, all_nodes,
                         policy_index=policy_index)


def obj_main():
    # Make stdout able to write escaped bytes
    sys.stdout = codecs.getwriter("utf-8")(
        sys.stdout.detach(), errors='surrogateescape')

    parser = OptionParser('%prog [options] OBJECT_FILE')
    parser.add_option(
        '-n', '--no-check-etag', default=True,
        action="store_false", dest="check_etag",
        help="Don't verify file contents against stored etag")
    parser.add_option(
        '-d', '--swift-dir', default='/etc/swift', dest='swift_dir',
        help="Pass location of swift directory")
    parser.add_option(
        '--drop-prefixes', default=False, action="store_true",
        help="When outputting metadata, drop the per-section common prefixes")
    parser.add_option(
        '-P', '--policy-name', dest='policy_name',
        help="Specify storage policy name")

    options, args = parser.parse_args()

    if len(args) != 1:
        sys.exit(parser.print_help())

    if set_swift_dir(options.swift_dir):
        reload_storage_policies()

    try:
        print_obj(*args, **vars(options))
    except InfoSystemExit:
        sys.exit(1)


def run_print_info(db_type, args, opts):
    try:
        print_info(db_type, *args, **opts)
    except InfoSystemExit:
        sys.exit(1)
    except (sqlite3.OperationalError, LockTimeout) as e:
        if not opts.get('stale_reads_ok'):
            opts['stale_reads_ok'] = True
            print('WARNING: Possibly Stale Data')
            run_print_info(db_type, args, opts)
            sys.exit(2)
        else:
            print('%s info failed: %s' % (db_type.title(), e))
            sys.exit(1)


def container_main():
    parser = OptionParser('%prog [options] CONTAINER_DB_FILE')
    parser.add_option(
        '-d', '--swift-dir', default='/etc/swift',
        help="Pass location of swift directory")
    parser.add_option(
        '--drop-prefixes', default=False, action="store_true",
        help="When outputting metadata, drop the per-section common prefixes")
    parser.add_option(
        '-v', '--verbose', default=False, action="store_true",
        help="Show all shard ranges. By default, only the number of shard "
             "ranges is displayed if there are many shards.")
    parser.add_option(
        '--sync', '-s', default=False, action="store_true",
        help="Output the contents of the incoming/outging sync tables")

    options, args = parser.parse_args()

    if len(args) != 1:
        sys.exit(parser.print_help())

    run_print_info('container', args, vars(options))


def account_main():
    parser = OptionParser('%prog [options] ACCOUNT_DB_FILE')
    parser.add_option(
        '-d', '--swift-dir', default='/etc/swift',
        help="Pass location of swift directory")
    parser.add_option(
        '--drop-prefixes', default=False, action="store_true",
        help="When outputting metadata, drop the per-section common prefixes")
    parser.add_option(
        '--sync', '-s', default=False, action="store_true",
        help="Output the contents of the incoming/outging sync tables")

    options, args = parser.parse_args()

    if len(args) != 1:
        sys.exit(parser.print_help())

    run_print_info('account', args, vars(options))



# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import unittest

from swift.common import digest
from test.debug_logger import debug_logger


class TestDigestUtils(unittest.TestCase):
    """Tests for swift.common.middleware.digest """
    def setUp(self):
        self.logger = debug_logger('test_digest_utils')

    def test_get_hmac(self):
        self.assertEqual(
            digest.get_hmac('GET', '/path', 1, 'abc'),
            'b17f6ff8da0e251737aa9e3ee69a881e3e092e2f')

    def test_get_hmac_ip_range(self):
        self.assertEqual(
            digest.get_hmac('GET', '/path', 1, 'abc', ip_range='127.0.0.1'),
            'b30dde4d2b8562b8496466c3b46b2b9ac5054461')

    def test_get_hmac_ip_range_non_binary_type(self):
        self.assertEqual(
            digest.get_hmac(
                u'GET', u'/path', 1, u'abc', ip_range=u'127.0.0.1'),
            'b30dde4d2b8562b8496466c3b46b2b9ac5054461')

    def test_get_hmac_digest(self):
        self.assertEqual(
            digest.get_hmac(u'GET', u'/path', 1, u'abc', digest='sha256'),
            '64c5558394f86b042ce1e929b34907abd9d0a57f3e20cd3f93cffd83de0206a7')
        self.assertEqual(
            digest.get_hmac(
                u'GET', u'/path', 1, u'abc', digest=hashlib.sha256),
            '64c5558394f86b042ce1e929b34907abd9d0a57f3e20cd3f93cffd83de0206a7')

        self.assertEqual(
            digest.get_hmac(u'GET', u'/path', 1, u'abc', digest='sha512'),
            '7e95af818aec1b69b53fc2cb6d69456ec64ebda6c17b8fc8b7303b78acc8ca'
            '14fc4aed96c1614a8e9d6ff45a6237711d8be294cda679624825d79aa6959b'
            '5229')
        self.assertEqual(
            digest.get_hmac(
                u'GET', u'/path', 1, u'abc', digest=hashlib.sha512),
            '7e95af818aec1b69b53fc2cb6d69456ec64ebda6c17b8fc8b7303b78acc8ca'
            '14fc4aed96c1614a8e9d6ff45a6237711d8be294cda679624825d79aa6959b'
            '5229')

    def test_extract_digest_and_algorithm(self):
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'b17f6ff8da0e251737aa9e3ee69a881e3e092e2f'),
            ('sha1', 'b17f6ff8da0e251737aa9e3ee69a881e3e092e2f'))
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'sha1:sw3eTSuFYrhJZGbDtGsrmsUFRGE='),
            ('sha1', 'b30dde4d2b8562b8496466c3b46b2b9ac5054461'))
        # also good with '=' stripped
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'sha1:sw3eTSuFYrhJZGbDtGsrmsUFRGE'),
            ('sha1', 'b30dde4d2b8562b8496466c3b46b2b9ac5054461'))

        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'b963712313cd4236696fb4c4cf11fc56'
                'ff4158e0bcbf1d4424df147783fd1045'),
            ('sha256', 'b963712313cd4236696fb4c4cf11fc56'
                       'ff4158e0bcbf1d4424df147783fd1045'))
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'sha256:uWNxIxPNQjZpb7TEzxH8Vv9BWOC8vx1EJN8Ud4P9EEU='),
            ('sha256', 'b963712313cd4236696fb4c4cf11fc56'
                       'ff4158e0bcbf1d4424df147783fd1045'))
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'sha256:uWNxIxPNQjZpb7TEzxH8Vv9BWOC8vx1EJN8Ud4P9EEU'),
            ('sha256', 'b963712313cd4236696fb4c4cf11fc56'
                       'ff4158e0bcbf1d4424df147783fd1045'))

        self.assertEqual(
            digest.extract_digest_and_algorithm(
                '26df3d9d59da574d6f8d359cb2620b1b'
                '86737215c38c412dfee0a410acea1ac4'
                '285ad0c37229ca74e715c443979da17d'
                '3d77a97d2ac79cc5e395b05bfa4bdd30'),
            ('sha512', '26df3d9d59da574d6f8d359cb2620b1b'
                       '86737215c38c412dfee0a410acea1ac4'
                       '285ad0c37229ca74e715c443979da17d'
                       '3d77a97d2ac79cc5e395b05bfa4bdd30'))
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'sha512:Jt89nVnaV01vjTWcsmILG4ZzchXDjEEt/uCkEKzq'
                'GsQoWtDDcinKdOcVxEOXnaF9PXepfSrHnMXjlbBb+kvdMA=='),
            ('sha512', '26df3d9d59da574d6f8d359cb2620b1b'
                       '86737215c38c412dfee0a410acea1ac4'
                       '285ad0c37229ca74e715c443979da17d'
                       '3d77a97d2ac79cc5e395b05bfa4bdd30'))
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'sha512:Jt89nVnaV01vjTWcsmILG4ZzchXDjEEt_uCkEKzq'
                'GsQoWtDDcinKdOcVxEOXnaF9PXepfSrHnMXjlbBb-kvdMA'),
            ('sha512', '26df3d9d59da574d6f8d359cb2620b1b'
                       '86737215c38c412dfee0a410acea1ac4'
                       '285ad0c37229ca74e715c443979da17d'
                       '3d77a97d2ac79cc5e395b05bfa4bdd30'))

        with self.assertRaises(ValueError):
            digest.extract_digest_and_algorithm('')
        with self.assertRaises(ValueError):
            digest.extract_digest_and_algorithm(
                'exactly_forty_chars_but_not_hex_encoded!')
        # Too short (md5)
        with self.assertRaises(ValueError):
            digest.extract_digest_and_algorithm(
                'd41d8cd98f00b204e9800998ecf8427e')
        # but you can slip it in via the prefix notation!
        self.assertEqual(
            digest.extract_digest_and_algorithm(
                'md5:1B2M2Y8AsgTpgAmY7PhCfg'),
            ('md5', 'd41d8cd98f00b204e9800998ecf8427e'))

    def test_get_allowed_digests(self):
        # start with defaults
        allowed, deprecated = digest.get_allowed_digests(
            ''.split(), self.logger)
        self.assertEqual(allowed, {'sha256', 'sha512', 'sha1'})
        self.assertEqual(deprecated, {'sha1'})
        warning_lines = self.logger.get_lines_for_level('warning')
        expected_warning_line = (
            'The following digest algorithms are allowed by default but '
            'deprecated: sha1. Support will be disabled by default in a '
            'future release, and later removed entirely.')
        self.assertIn(expected_warning_line, warning_lines)
        self.logger.clear()

        # now with a subset
        allowed, deprecated = digest.get_allowed_digests(
            'sha1 sha256'.split(), self.logger)
        self.assertEqual(allowed, {'sha256', 'sha1'})
        self.assertEqual(deprecated, {'sha1'})
        warning_lines = self.logger.get_lines_for_level('warning')
        expected_warning_line = (
            'The following digest algorithms are configured but '
            'deprecated: sha1. Support will be removed in a future release.')
        self.assertIn(expected_warning_line, warning_lines)
        self.logger.clear()

        # Now also with an unsupported digest
        allowed, deprecated = digest.get_allowed_digests(
            'sha1 sha256 md5'.split(), self.logger)
        self.assertEqual(allowed, {'sha256', 'sha1'})
        self.assertEqual(deprecated, {'sha1'})
        warning_lines = self.logger.get_lines_for_level('warning')
        self.assertIn(expected_warning_line, warning_lines)
        expected_unsupported_warning_line = (
            'The following digest algorithms are configured but not '
            'supported: md5')
        self.assertIn(expected_unsupported_warning_line, warning_lines)
        self.logger.clear()

        # Now with no deprecated digests
        allowed, deprecated = digest.get_allowed_digests(
            'sha256 sha512'.split(), self.logger)
        self.assertEqual(allowed, {'sha256', 'sha512'})
        self.assertEqual(deprecated, set())
        warning_lines = self.logger.get_lines_for_level('warning')
        self.assertFalse(warning_lines)
        self.logger.clear()

        # no valid digest
        # Now also with an unsupported digest
        with self.assertRaises(ValueError):
            digest.get_allowed_digests(['md5'], self.logger)
        warning_lines = self.logger.get_lines_for_level('warning')
        self.assertIn(expected_unsupported_warning_line, warning_lines)



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import unquote

from swift.account.utils import account_listing_response
from swift.common.middleware.acl import parse_acl, format_acl
from swift.common.utils import public
from swift.common.constraints import check_metadata
from swift.common.http import HTTP_NOT_FOUND, HTTP_GONE
from swift.proxy.controllers.base import Controller, clear_info_cache, \
    set_info_cache, NodeIter
from swift.common.middleware import listing_formats
from swift.common.swob import HTTPBadRequest, HTTPMethodNotAllowed
from swift.common.request_helpers import get_sys_meta_prefix


class AccountController(Controller):
    """WSGI controller for account requests"""
    server_type = 'Account'

    def __init__(self, app, account_name, **kwargs):
        super(AccountController, self).__init__(app)
        self.account_name = unquote(account_name)
        if not self.app.allow_account_management:
            self.allowed_methods.remove('PUT')
            self.allowed_methods.remove('DELETE')

    def add_acls_from_sys_metadata(self, resp):
        if resp.environ['REQUEST_METHOD'] in ('HEAD', 'GET', 'PUT', 'POST'):
            prefix = get_sys_meta_prefix('account') + 'core-'
            name = 'access-control'
            (extname, intname) = ('x-account-' + name, prefix + name)
            acl_dict = parse_acl(version=2, data=resp.headers.pop(intname))
            if acl_dict:  # treat empty dict as empty header
                resp.headers[extname] = format_acl(
                    version=2, acl_dict=acl_dict)

    def GETorHEAD(self, req):
        """Handler for HTTP GET/HEAD requests."""
        length_limit = self.get_name_length_limit()
        if len(self.account_name) > length_limit:
            resp = HTTPBadRequest(request=req)
            resp.body = b'Account name length of %d longer than %d' % \
                        (len(self.account_name), length_limit)
            # Don't cache this. We know the account doesn't exist because
            # the name is bad; we don't need to cache that because it's
            # really cheap to recompute.
            return resp

        partition = self.app.account_ring.get_part(self.account_name)
        concurrency = self.app.account_ring.replica_count \
            if self.app.get_policy_options(None).concurrent_gets else 1
        node_iter = NodeIter(
            'account', self.app, self.app.account_ring, partition,
            self.logger, req)
        params = req.params
        params['format'] = 'json'
        req.params = params
        resp = self.GETorHEAD_base(
            req, 'Account', node_iter, partition,
            req.swift_entity_path.rstrip('/'), concurrency)
        if resp.status_int == HTTP_NOT_FOUND:
            if resp.headers.get('X-Account-Status', '').lower() == 'deleted':
                resp.status = HTTP_GONE
            elif self.app.account_autocreate:
                # This is kind of a lie; we pretend like the account is
                # there, but it's not. We'll create it as soon as something
                # tries to write to it, but we don't need databases on disk
                # to tell us that nothing's there.
                #
                # We set a header so that certain consumers can tell it's a
                # fake listing. The important one is the PUT of a container
                # to an autocreate account; the proxy checks to see if the
                # account exists before actually performing the PUT and
                # creates the account if necessary. If we feed it a perfect
                # lie, it'll just try to create the container without
                # creating the account, and that'll fail.
                resp = account_listing_response(
                    self.account_name, req,
                    listing_formats.get_listing_content_type(req))
                resp.headers['X-Backend-Fake-Account-Listing'] = 'yes'

        # Cache this. We just made a request to a storage node and got
        # up-to-date information for the account.
        resp.headers['X-Backend-Recheck-Account-Existence'] = str(
            self.app.recheck_account_existence)
        set_info_cache(req.environ, self.account_name, None, resp)

        if req.environ.get('swift_owner'):
            self.add_acls_from_sys_metadata(resp)
        else:
            for header in self.app.swift_owner_headers:
                resp.headers.pop(header, None)
        return resp

    @public
    def PUT(self, req):
        """HTTP PUT request handler."""
        if not self.app.allow_account_management:
            return HTTPMethodNotAllowed(
                request=req,
                headers={'Allow': ', '.join(self.allowed_methods)})
        error_response = check_metadata(req, 'account')
        if error_response:
            return error_response
        length_limit = self.get_name_length_limit()
        if len(self.account_name) > length_limit:
            resp = HTTPBadRequest(request=req)
            resp.body = b'Account name length of %d longer than %d' % \
                        (len(self.account_name), length_limit)
            return resp
        account_partition, accounts = \
            self.app.account_ring.get_nodes(self.account_name)
        headers = self.generate_request_headers(req, transfer=True)
        clear_info_cache(req.environ, self.account_name)
        resp = self.make_requests(
            req, self.app.account_ring, account_partition, 'PUT',
            req.swift_entity_path, [headers] * len(accounts))
        self.add_acls_from_sys_metadata(resp)
        return resp

    @public
    def POST(self, req):
        """HTTP POST request handler."""
        length_limit = self.get_name_length_limit()
        if len(self.account_name) > length_limit:
            resp = HTTPBadRequest(request=req)
            resp.body = b'Account name length of %d longer than %d' % \
                        (len(self.account_name), length_limit)
            return resp
        error_response = check_metadata(req, 'account')
        if error_response:
            return error_response
        account_partition, accounts = \
            self.app.account_ring.get_nodes(self.account_name)
        headers = self.generate_request_headers(req, transfer=True)
        clear_info_cache(req.environ, self.account_name)
        resp = self.make_requests(
            req, self.app.account_ring, account_partition, 'POST',
            req.swift_entity_path, [headers] * len(accounts))
        if resp.status_int == HTTP_NOT_FOUND and self.app.account_autocreate:
            self.autocreate_account(req, self.account_name)
            resp = self.make_requests(
                req, self.app.account_ring, account_partition, 'POST',
                req.swift_entity_path, [headers] * len(accounts))
        self.add_acls_from_sys_metadata(resp)
        return resp

    @public
    def DELETE(self, req):
        """HTTP DELETE request handler."""
        # Extra safety in case someone typos a query string for an
        # account-level DELETE request that was really meant to be caught by
        # some middleware.
        if req.query_string:
            return HTTPBadRequest(request=req)
        if not self.app.allow_account_management:
            return HTTPMethodNotAllowed(
                request=req,
                headers={'Allow': ', '.join(self.allowed_methods)})
        account_partition, accounts = \
            self.app.account_ring.get_nodes(self.account_name)
        headers = self.generate_request_headers(req)
        clear_info_cache(req.environ, self.account_name)
        resp = self.make_requests(
            req, self.app.account_ring, account_partition, 'DELETE',
            req.swift_entity_path, [headers] * len(accounts))
        return resp



# Copyright (c) 2010-2020 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock
import unittest

from swift.common import swob
from swift.common.middleware import etag_quoter
from swift.proxy.controllers.base import get_cache_key

from test.unit.common.middleware.helpers import FakeSwift


def set_info_cache(req, cache_data, account, container=None):
    req.environ.setdefault('swift.infocache', {})[
        get_cache_key(account, container)] = cache_data


class TestEtagQuoter(unittest.TestCase):
    def get_mw(self, conf, etag='unquoted-etag', path=None):
        if path is None:
            path = '/v1/AUTH_acc/con/some/path/to/obj'
        app = FakeSwift()
        hdrs = {} if etag is None else {'ETag': etag}
        app.register('GET', path, swob.HTTPOk, hdrs)
        return etag_quoter.filter_factory({}, **conf)(app)

    @mock.patch('swift.common.middleware.etag_quoter.register_swift_info')
    def test_swift_info(self, mock_register):
        self.get_mw({})
        self.assertEqual(mock_register.mock_calls, [
            mock.call('etag_quoter', enable_by_default=False)])
        mock_register.reset_mock()

        self.get_mw({'enable_by_default': '1'})
        self.assertEqual(mock_register.mock_calls, [
            mock.call('etag_quoter', enable_by_default=True)])
        mock_register.reset_mock()

        self.get_mw({'enable_by_default': 'no'})
        self.assertEqual(mock_register.mock_calls, [
            mock.call('etag_quoter', enable_by_default=False)])

    def test_account_on_overrides_cluster_off(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': '1'},
        }, 'AUTH_acc')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {},
        }, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 'false'}))
        self.assertEqual(resp.headers['ETag'], '"unquoted-etag"')

    def test_account_off_overrides_cluster_on(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': 'no'},
        }, 'AUTH_acc')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {},
        }, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 'yes'}))
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')

    def test_container_on_overrides_cluster_off(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {},
        }, 'AUTH_acc')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': 't'},
        }, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 'false'}))
        self.assertEqual(resp.headers['ETag'], '"unquoted-etag"')

    def test_container_off_overrides_cluster_on(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {},
        }, 'AUTH_acc')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': '0'},
        }, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 'yes'}))
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')

    def test_container_on_overrides_account_off(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': 'no'},
        }, 'AUTH_acc')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': 't'},
        }, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({}))
        self.assertEqual(resp.headers['ETag'], '"unquoted-etag"')

    def test_container_off_overrides_account_on(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': 'yes'},
        }, 'AUTH_acc')
        set_info_cache(req, {
            'status': 200,
            'sysmeta': {'rfc-compliant-etags': 'false'},
        }, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({}))
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')

    def test_cluster_wide(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 't'}))
        self.assertEqual(resp.headers['ETag'], '"unquoted-etag"')

    def test_already_valid(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 't'},
                                            '"quoted-etag"'))
        self.assertEqual(resp.headers['ETag'], '"quoted-etag"')

    def test_already_weak_but_valid(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 't'},
                                            'W/"weak-etag"'))
        self.assertEqual(resp.headers['ETag'], 'W/"weak-etag"')

    def test_only_half_valid(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 't'},
                                            '"weird-etag'))
        self.assertEqual(resp.headers['ETag'], '""weird-etag"')

    def test_no_etag(self):
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(self.get_mw({'enable_by_default': 't'},
                                            etag=None))
        self.assertNotIn('ETag', resp.headers)

    def test_non_swift_path(self):
        path = '/some/other/location/entirely'
        req = swob.Request.blank(path)
        resp = req.get_response(self.get_mw({'enable_by_default': 't'},
                                            path=path))
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')

    def test_non_object_request(self):
        path = '/v1/AUTH_acc/con'
        req = swob.Request.blank(path)
        resp = req.get_response(self.get_mw({'enable_by_default': 't'},
                                            path=path))
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')

    def test_no_container_info(self):
        mw = self.get_mw({'enable_by_default': 't'})
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        mw.app.register('HEAD', '/v1/AUTH_acc/con',
                        swob.HTTPServiceUnavailable, {})
        resp = req.get_response(mw)
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')
        set_info_cache(req, {'status': 404, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(mw)
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(mw)
        self.assertEqual(resp.headers['ETag'], '"unquoted-etag"')

    def test_no_account_info(self):
        mw = self.get_mw({'enable_by_default': 't'})
        req = swob.Request.blank('/v1/AUTH_acc/con/some/path/to/obj')
        mw.app.register('HEAD', '/v1/AUTH_acc',
                        swob.HTTPServiceUnavailable, {})
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc', 'con')
        resp = req.get_response(mw)
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')
        set_info_cache(req, {'status': 404, 'sysmeta': {}}, 'AUTH_acc')
        resp = req.get_response(mw)
        self.assertEqual(resp.headers['ETag'], 'unquoted-etag')
        set_info_cache(req, {'status': 200, 'sysmeta': {}}, 'AUTH_acc')
        resp = req.get_response(mw)
        self.assertEqual(resp.headers['ETag'], '"unquoted-etag"')



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Object Server for Swift """

import pickle  # nosec: B403
from urllib.parse import unquote
import json
import os
import multiprocessing
import sys
import time
import traceback
import socket

from eventlet import sleep, wsgi, Timeout, tpool
from eventlet.greenthread import spawn

from swift.common.utils import public, get_logger, \
    config_true_value, config_percent_value, \
    replication, normalize_delete_at_timestamp, \
    get_log_line, Timestamp, parse_mime_headers, \
    iter_multipart_mime_documents, extract_swift_bytes, safe_json_loads, \
    config_auto_int_value, split_path, get_redirect_data, \
    normalize_timestamp, md5, parse_options, CooperativeIterator
from swift.common.bufferedhttp import http_connect
from swift.common.constraints import check_object_creation, \
    valid_timestamp, check_utf8, AUTO_CREATE_ACCOUNT_PREFIX
from swift.common.exceptions import ConnectionTimeout, DiskFileQuarantined, \
    DiskFileNotExist, DiskFileCollision, DiskFileNoSpace, DiskFileDeleted, \
    DiskFileDeviceUnavailable, DiskFileExpired, ChunkReadTimeout, \
    ChunkReadError, DiskFileXattrNotSupported, DiskFileStateChanged
from swift.common.request_helpers import resolve_ignore_range_header, \
    OBJECT_SYSMETA_CONTAINER_UPDATE_OVERRIDE_PREFIX
from swift.obj import ssync_receiver, expirer
from swift.common.http import is_success, HTTP_MOVED_PERMANENTLY
from swift.common.base_storage_server import BaseStorageServer, \
    timing_stats, labeled_timing_stats
from swift.common.header_key_dict import HeaderKeyDict
from swift.common.request_helpers import get_name_and_placement, \
    is_user_meta, is_sys_or_user_meta, is_object_transient_sysmeta, \
    resolve_etag_is_at_header, is_sys_meta, validate_internal_obj, \
    is_backend_open_expired
from swift.common.statsd_client import get_labeled_statsd_client
from swift.common.swob import HTTPAccepted, HTTPBadRequest, HTTPCreated, \
    HTTPInternalServerError, HTTPNoContent, HTTPNotFound, \
    HTTPPreconditionFailed, HTTPRequestTimeout, HTTPUnprocessableEntity, \
    HTTPClientDisconnect, HTTPMethodNotAllowed, Request, Response, \
    HTTPInsufficientStorage, HTTPForbidden, HTTPException, HTTPConflict, \
    HTTPServerError, bytes_to_wsgi, wsgi_to_bytes, wsgi_to_str, \
    normalize_etag, HTTPServiceUnavailable
from swift.common.wsgi import run_wsgi
from swift.obj.diskfile import RESERVED_DATAFILE_META, DiskFileRouter
from swift.obj.expirer import build_task_obj, embed_expirer_bytes_in_ctype, \
    X_DELETE_TYPE


LABELED_METRIC_NAME = 'swift_object_server_request_timing'


def iter_mime_headers_and_bodies(wsgi_input, mime_boundary, read_chunk_size):
    mime_documents_iter = iter_multipart_mime_documents(
        wsgi_input, mime_boundary, read_chunk_size)

    for file_like in mime_documents_iter:
        hdrs = parse_mime_headers(file_like)
        yield (hdrs, file_like)


def drain(file_like, read_size, timeout):
    """
    Read and discard any bytes from file_like.

    :param file_like: file-like object to read from
    :param read_size: how big a chunk to read at a time
    :param timeout: how long to wait for a read (use None for no timeout)

    :raises ChunkReadTimeout: if no chunk was read in time
    """

    while True:
        with ChunkReadTimeout(timeout):
            chunk = file_like.read(read_size)
            if not chunk:
                break


def get_obj_name_and_placement(request):
    """
    Split and validate path for an object.

    :param request: a swob request

    :returns: a tuple of path parts and storage policy
    """
    device, partition, account, container, obj, policy = \
        get_name_and_placement(request, 5, 5, True)
    validate_internal_obj(account, container, obj)
    return device, partition, account, container, obj, policy


def _make_backend_fragments_header(fragments):
    if fragments:
        result = {}
        for ts, frag_list in fragments.items():
            result[ts.internal] = frag_list
        return json.dumps(result)
    return None


class EventletPlungerString(bytes):
    """
    Eventlet won't send headers until it's accumulated at least
    eventlet.wsgi.MINIMUM_CHUNK_SIZE bytes or the app iter is exhausted.
    If we want to send the response body behind Eventlet's back, perhaps
    with some zero-copy wizardry, then we have to unclog the plumbing in
    eventlet.wsgi to force the headers out, so we use an
    EventletPlungerString to empty out all of Eventlet's buffers.
    """
    def __len__(self):
        return wsgi.MINIMUM_CHUNK_SIZE + 1


class ObjectController(BaseStorageServer):
    """Implements the WSGI application for the Swift Object Server."""

    server_type = 'object-server'

    def __init__(self, conf, logger=None):
        """
        Creates a new WSGI application for the Swift Object Server. An
        example configuration is given at
        <source-dir>/etc/object-server.conf-sample or
        /etc/swift/object-server.conf-sample.
        """
        super(ObjectController, self).__init__(conf)
        self.logger = logger or get_logger(conf, log_route='object-server')
        self.statsd = get_labeled_statsd_client(conf, self.logger.logger)
        self.node_timeout = float(conf.get('node_timeout', 3))
        self.container_update_timeout = float(
            conf.get('container_update_timeout', 1))
        self.conn_timeout = float(conf.get('conn_timeout', 0.5))
        self.client_timeout = float(conf.get('client_timeout', 60))
        self.disk_chunk_size = int(conf.get('disk_chunk_size', 65536))
        self.network_chunk_size = int(conf.get('network_chunk_size', 65536))
        self.log_requests = config_true_value(conf.get('log_requests', 'true'))
        self.max_upload_time = int(conf.get('max_upload_time', 86400))
        self.slow = int(conf.get('slow', 0))
        self.keep_cache_private = \
            config_true_value(conf.get('keep_cache_private', 'false'))
        self.keep_cache_slo_manifest = \
            config_true_value(conf.get('keep_cache_slo_manifest', 'false'))
        self.cooperative_period = int(conf.get("cooperative_period", 0))
        self.etag_validate_frac = config_percent_value(
            conf.get("etag_validate_pct", 100))

        default_allowed_headers = '''
            content-disposition,
            content-encoding,
            x-delete-at,
            x-object-manifest,
            x-static-large-object,
            cache-control,
            content-language,
            expires,
            x-robots-tag
        '''
        extra_allowed_headers = [
            header.strip().lower() for header in conf.get(
                'allowed_headers', default_allowed_headers).split(',')
            if header.strip()
        ]
        self.allowed_headers = set()
        for header in extra_allowed_headers:
            if header not in RESERVED_DATAFILE_META:
                self.allowed_headers.add(header)

        self.auto_create_account_prefix = AUTO_CREATE_ACCOUNT_PREFIX
        self.expirer_config = expirer.ExpirerConfig(conf, logger=self.logger)
        # Initialization was successful, so now apply the network chunk size
        # parameter as the default read / write buffer size for the network
        # sockets.
        #
        # NOTE WELL: This is a class setting, so until we get set this on a
        # per-connection basis, this affects reading and writing on ALL
        # sockets, those between the proxy servers and external clients, and
        # those between the proxy servers and the other internal servers.
        #
        # ** Because the primary motivation for this is to optimize how data
        # is written back to the proxy server, we could use the value from the
        # disk_chunk_size parameter. However, it affects all created sockets
        # using this class so we have chosen to tie it to the
        # network_chunk_size parameter value instead.
        # if six.PY2:
        #    socket._fileobject.default_bufsize = self.network_chunk_size
        # TODO: find a way to enable similar functionality in py3

        # Provide further setup specific to an object server implementation.
        self.setup(conf)

    def setup(self, conf):
        """
        Implementation specific setup. This method is called at the very end
        by the constructor to allow a specific implementation to modify
        existing attributes or add its own attributes.

        :param conf: WSGI configuration parameter
        """

        # Common on-disk hierarchy shared across account, container and object
        # servers.
        self._diskfile_router = DiskFileRouter(conf, self.logger)
        # This is populated by global_conf_callback way below as the semaphore
        # is shared by all workers.
        if 'replication_semaphore' in conf:
            # The value was put in a list so it could get past paste
            self.replication_semaphore = conf['replication_semaphore'][0]
        else:
            self.replication_semaphore = None
        self.replication_failure_threshold = int(
            conf.get('replication_failure_threshold') or 100)
        self.replication_failure_ratio = float(
            conf.get('replication_failure_ratio') or 1.0)

        servers_per_port = int(conf.get('servers_per_port', '0') or 0)
        if servers_per_port:
            # The typical servers-per-port deployment also uses one port per
            # disk, so you really get N servers per disk. In that case,
            # having a pool of 20 threads per server per disk is far too
            # much. For example, given a 60-disk chassis and 4 servers per
            # disk, the default configuration will give us 21 threads per
            # server (the main thread plus the twenty tpool threads), for a
            # total of around 60 * 21 * 4 = 5040 threads. This is clearly
            # too high.
            #
            # Instead, we use a tpool size of 1, giving us 2 threads per
            # process. In the example above, that's 60 * 2 * 4 = 480
            # threads, which is reasonable since there are 240 processes.
            default_tpool_size = 1
        else:
            # If we're not using servers-per-port, then leave the tpool size
            # alone. The default (20) is typically good enough for one
            # object server handling requests for many disks.
            default_tpool_size = None

        tpool_size = config_auto_int_value(
            conf.get('eventlet_tpool_num_threads'),
            default_tpool_size)

        if tpool_size:
            tpool.set_num_threads(tpool_size)

    def get_diskfile(self, device, partition, account, container, obj,
                     policy, **kwargs):
        """
        Utility method for instantiating a DiskFile object supporting a given
        REST API.

        An implementation of the object server that wants to use a different
        DiskFile class would simply over-ride this method to provide that
        behavior.
        """
        return self._diskfile_router[policy].get_diskfile(
            device, partition, account, container, obj, policy, **kwargs)

    def async_update(self, op, account, container, obj, host, partition,
                     contdevice, headers_out, objdevice, policy,
                     logger_thread_locals=None, container_path=None,
                     db_state=None):
        """
        Sends or saves an async update.

        :param op: operation performed (ex: 'PUT', or 'DELETE')
        :param account: account name for the object
        :param container: container name for the object
        :param obj: object name
        :param host: host that the container is on
        :param partition: partition that the container is on
        :param contdevice: device name that the container is on
        :param headers_out: dictionary of headers to send in the container
                            request
        :param objdevice: device name that the object is in
        :param policy: the associated BaseStoragePolicy instance
        :param logger_thread_locals: The thread local values to be set on the
                                     self.logger to retain transaction
                                     logging information.
        :param container_path: optional path in the form `<account/container>`
            to which the update should be sent. If given this path will be used
            instead of constructing a path from the ``account`` and
            ``container`` params.
        :param db_state: The current database state of the container as
            supplied to us by the proxy.
        """
        if logger_thread_locals:
            self.logger.thread_locals = logger_thread_locals
        headers_out['user-agent'] = 'object-server %s' % os.getpid()
        if container_path:
            # use explicitly specified container path
            full_path = '/%s/%s' % (container_path, obj)
        else:
            full_path = '/%s/%s/%s' % (account, container, obj)

        redirect_data = None
        if all([host, partition, contdevice]):
            try:
                with ConnectionTimeout(self.conn_timeout):
                    ip, port = host.rsplit(':', 1)
                    conn = http_connect(ip, port, contdevice, partition, op,
                                        full_path, headers_out)
                with Timeout(self.node_timeout):
                    response = conn.getresponse()
                    response.read()
                if is_success(response.status):
                    return

                if response.status == HTTP_MOVED_PERMANENTLY:
                    try:
                        redirect_data = get_redirect_data(response)
                    except ValueError as err:
                        self.logger.error(
                            'Container update failed for %r; problem with '
                            'redirect location: %s' % (obj, err))
                else:
                    self.logger.error(
                        'ERROR Container update failed '
                        '(saving for async update later): %(status)d '
                        'response from %(ip)s:%(port)s/%(dev)s',
                        {'status': response.status, 'ip': ip, 'port': port,
                         'dev': contdevice})
            except (Exception, Timeout):
                self.logger.exception(
                    'ERROR container update failed with '
                    '%(ip)s:%(port)s/%(dev)s (saving for async update later)',
                    {'ip': ip, 'port': port, 'dev': contdevice})
        data = {'op': op, 'account': account, 'container': container,
                'obj': obj, 'headers': headers_out, 'db_state': db_state}
        if redirect_data:
            self.logger.debug(
                'Update to %(path)s redirected to %(redirect)s',
                {'path': full_path, 'redirect': redirect_data[0]})
            container_path = redirect_data[0]
        if container_path:
            data['container_path'] = container_path
        timestamp = headers_out.get('x-meta-timestamp',
                                    headers_out.get('x-timestamp'))
        self._diskfile_router[policy].pickle_async_update(
            objdevice, account, container, obj, data, timestamp, policy)

    def container_update(self, op, account, container, obj, request,
                         headers_out, objdevice, policy):
        """
        Update the container when objects are updated.

        :param op: operation performed (ex: 'PUT', or 'DELETE')
        :param account: account name for the object
        :param container: container name for the object
        :param obj: object name
        :param request: the original request object driving the update
        :param headers_out: dictionary of headers to send in the container
                            request(s)
        :param objdevice: device name that the object is in
        :param policy:  the BaseStoragePolicy instance
        """
        headers_in = request.headers
        conthosts = [h.strip() for h in
                     headers_in.get('X-Container-Host', '').split(',')]
        contdevices = [d.strip() for d in
                       headers_in.get('X-Container-Device', '').split(',')]
        contpartition = headers_in.get('X-Container-Partition', '')
        contdbstate = headers_in.get('X-Container-Root-Db-State')

        if len(conthosts) != len(contdevices):
            # This shouldn't happen unless there's a bug in the proxy,
            # but if there is, we want to know about it.
            self.logger.error(
                'ERROR Container update failed: different '
                'numbers of hosts and devices in request: '
                '"%(hosts)s" vs "%(devices)s"', {
                    'hosts': headers_in.get('X-Container-Host', ''),
                    'devices': headers_in.get('X-Container-Device', '')})
            return

        contpath = headers_in.get('X-Backend-Quoted-Container-Path')
        if contpath:
            contpath = unquote(contpath)
        else:
            contpath = headers_in.get('X-Backend-Container-Path')

        if contpath:
            try:
                # TODO: this is very late in request handling to be validating
                # a header - if we did *not* check and the header was bad
                # presumably the update would fail and we would fall back to an
                # async update to the root container, which might be best
                # course of action rather than aborting update altogether?
                split_path('/' + contpath, minsegs=2, maxsegs=2)
            except ValueError:
                self.logger.error(
                    "Invalid X-Backend-Container-Path, should be of the form "
                    "'account/container' but got %r." % contpath)
                # fall back to updating root container
                contpath = None

        if contpartition:
            # In py3, zip() continues to work for our purposes... But when
            # we want to log an error, consumed items are not longer present
            # in the zip, making the logs useless for operators. So, list().
            updates = list(zip(conthosts, contdevices))
        else:
            updates = []

        headers_out['x-trans-id'] = headers_in.get('x-trans-id', '-')
        headers_out['referer'] = request.as_referer()
        headers_out['X-Backend-Storage-Policy-Index'] = int(policy)
        update_greenthreads = []
        for conthost, contdevice in updates:
            gt = spawn(self.async_update, op, account, container, obj,
                       conthost, contpartition, contdevice, headers_out,
                       objdevice, policy,
                       logger_thread_locals=self.logger.thread_locals,
                       container_path=contpath, db_state=contdbstate)
            update_greenthreads.append(gt)
        # Wait a little bit to see if the container updates are successful.
        # If we immediately return after firing off the greenthread above, then
        # we're more likely to confuse the end-user who does a listing right
        # after getting a successful response to the object create. The
        # `container_update_timeout` bounds the length of time we wait so that
        # one slow container server doesn't make the entire request lag.
        try:
            with Timeout(self.container_update_timeout):
                for gt in update_greenthreads:
                    gt.wait()
        except Timeout:
            # updates didn't go through, log it and return
            self.logger.debug(
                'Container update timeout (%.4fs) waiting for %s',
                self.container_update_timeout, updates)

    def delete_at_update(self, op, delete_at, account, container, obj,
                         request, objdevice, policy, extra_headers=None):
        """
        Update the expiring objects container when objects are updated.

        :param op: operation performed (ex: 'PUT', or 'DELETE')
        :param delete_at: scheduled delete in UNIX seconds, int
        :param account: account name for the object
        :param container: container name for the object
        :param obj: object name
        :param request: the original request driving the update
        :param objdevice: device name that the object is in
        :param policy: the BaseStoragePolicy instance (used for tmp dir)
        :param extra_headers: dict of additional headers for the update
        """
        if config_true_value(
                request.headers.get('x-backend-replication', 'f')):
            return

        delete_at = normalize_delete_at_timestamp(delete_at)

        headers_in = request.headers
        headers_out = HeaderKeyDict({
            # system accounts are always Policy-0
            'X-Backend-Storage-Policy-Index': 0,
            'x-timestamp': request.timestamp.internal,
            'x-trans-id': headers_in.get('x-trans-id', '-'),
            'referer': request.as_referer()})

        expiring_objects_account_name, delete_at_container = \
            self.expirer_config.get_expirer_account_and_container(
                delete_at, account, container, obj)
        if op != 'DELETE':
            hosts = headers_in.get('X-Delete-At-Host', None)
            if hosts is None:
                # If header is missing, no update needed as sufficient other
                # object servers should perform the required update.
                return

            proxy_delete_at_container = headers_in.get(
                'X-Delete-At-Container', None)
            if delete_at_container != proxy_delete_at_container:
                if not proxy_delete_at_container:
                    # We carry this warning around for pre-2013 proxies
                    self.logger.warning(
                        'X-Delete-At-Container header must be specified for '
                        'expiring objects background %s to work properly. '
                        'Making best guess as to the container name '
                        'for now.', op)
                    proxy_delete_at_container = delete_at_container
                else:
                    # Inconsistent configuration may lead to orphaned expirer
                    # task queue objects when X-Delete-At is updated, which can
                    # stick around for a whole reclaim age.
                    self.logger.debug(
                        'Proxy X-Delete-At-Container %r does not match '
                        'expected %r for current expirer_config.',
                        proxy_delete_at_container, delete_at_container)
                # it's not possible to say which is "more correct", this will
                # at least match the host/part/device
                delete_at_container = normalize_delete_at_timestamp(
                    proxy_delete_at_container)

            # new updates need to enqueue new x-delete-at
            partition = headers_in.get('X-Delete-At-Partition', None)
            contdevices = headers_in.get('X-Delete-At-Device', '')
            updates = [upd for upd in
                       zip((h.strip() for h in hosts.split(',')),
                           (c.strip() for c in contdevices.split(',')))
                       if all(upd) and partition]
            if not updates:
                updates = [(None, None)]
            headers_out['x-size'] = '0'
            headers_out['x-content-type'] = X_DELETE_TYPE
            headers_out['x-etag'] = 'd41d8cd98f00b204e9800998ecf8427e'
            if extra_headers:
                headers_out.update(extra_headers)
        else:
            if not config_true_value(
                request.headers.get(
                    'X-Backend-Clean-Expiring-Object-Queue', 't')):
                return
            # DELETE op always go directly to async_pending
            partition = None
            updates = [(None, None)]

        for host, contdevice in updates:
            self.async_update(
                op, expiring_objects_account_name, delete_at_container,
                build_task_obj(delete_at, account, container, obj),
                host, partition, contdevice, headers_out, objdevice,
                policy)

    def _make_timeout_reader(self, file_like):
        def timeout_reader():
            with ChunkReadTimeout(self.client_timeout):
                try:
                    return file_like.read(self.network_chunk_size)
                except (IOError, ValueError):
                    raise ChunkReadError
        return timeout_reader

    def _read_put_commit_message(self, mime_documents_iter):
        rcvd_commit = False
        try:
            with ChunkReadTimeout(self.client_timeout):
                commit_hdrs, commit_iter = next(mime_documents_iter)
                if commit_hdrs.get('X-Document', None) == "put commit":
                    rcvd_commit = True
            drain(commit_iter, self.network_chunk_size, self.client_timeout)
        except ChunkReadError:
            raise HTTPClientDisconnect()
        except ChunkReadTimeout:
            raise HTTPRequestTimeout()
        except StopIteration:
            raise HTTPBadRequest(body="couldn't find PUT commit MIME doc")
        return rcvd_commit

    def _read_metadata_footer(self, mime_documents_iter):
        try:
            with ChunkReadTimeout(self.client_timeout):
                footer_hdrs, footer_iter = next(mime_documents_iter)
        except ChunkReadError:
            raise HTTPClientDisconnect()
        except ChunkReadTimeout:
            raise HTTPRequestTimeout()
        except StopIteration:
            raise HTTPBadRequest(body="couldn't find footer MIME doc")
        return self._parse_footer(footer_hdrs, footer_iter)

    def _parse_footer(self, footer_hdrs, footer_iter):
        """
        Validate footer metadata and translate JSON body into HeaderKeyDict.
        """
        timeout_reader = self._make_timeout_reader(footer_iter)
        try:
            footer_body = b''.join(iter(timeout_reader, b''))
        except ChunkReadError:
            raise HTTPClientDisconnect()
        except ChunkReadTimeout:
            raise HTTPRequestTimeout()

        footer_md5 = footer_hdrs.get('Content-MD5')
        if not footer_md5:
            raise HTTPBadRequest(body="no Content-MD5 in footer")
        if footer_md5 != md5(footer_body, usedforsecurity=False).hexdigest():
            raise HTTPUnprocessableEntity(body="footer MD5 mismatch")

        try:
            return HeaderKeyDict(json.loads(footer_body))
        except ValueError:
            raise HTTPBadRequest("invalid JSON for footer doc")

    def _check_container_override(self, update_headers, metadata,
                                  footers=None):
        """
        Applies any overrides to the container update headers.

        Overrides may be in the x-object-sysmeta-container-update- namespace or
        the x-backend-container-update-override- namespace. The former is
        preferred and is used by proxy middlewares. The latter is historical
        but is still used with EC policy PUT requests; for backwards
        compatibility the header names used with EC policy requests have not
        been changed to the sysmeta namespace - that way the EC PUT path of a
        newer proxy will remain compatible with an object server that pre-dates
        the introduction of the x-object-sysmeta-container-update- namespace
        and vice-versa.

        :param update_headers: a dict of headers used in the container update
        :param metadata: a dict that may container override items
        :param footers: another dict that may container override items, at a
                        higher priority than metadata
        """
        footers = footers or {}
        # the order of this list is significant:
        # x-object-sysmeta-container-update-override-* headers take precedence
        # over x-backend-container-update-override-* headers
        override_prefixes = ['x-backend-container-update-override-',
                             OBJECT_SYSMETA_CONTAINER_UPDATE_OVERRIDE_PREFIX]
        for override_prefix in override_prefixes:
            for key, val in metadata.items():
                if key.lower().startswith(override_prefix):
                    override = key.lower().replace(override_prefix, 'x-')
                    update_headers[override] = val
            # apply x-backend-container-update-override* from footers *before*
            # x-object-sysmeta-container-update-override-* from headers
            for key, val in footers.items():
                if key.lower().startswith(override_prefix):
                    override = key.lower().replace(override_prefix, 'x-')
                    update_headers[override] = val

    def _conditional_delete_at_update(self, request, device, account,
                                      container, obj, policy, metadata,
                                      orig_delete_at, new_delete_at):
        if new_delete_at:
            extra_headers = {
                'x-content-type': embed_expirer_bytes_in_ctype(
                    X_DELETE_TYPE, metadata),
                'x-content-type-timestamp':
                metadata.get('X-Timestamp'),
            }
            self.delete_at_update(
                'PUT', new_delete_at, account, container, obj, request,
                device, policy, extra_headers)
        if orig_delete_at and orig_delete_at != new_delete_at:
            self.delete_at_update(
                'DELETE', orig_delete_at, account, container, obj,
                request, device, policy)

    @public
    @timing_stats()
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def POST(self, request, timing_stats_labels):
        """Handle HTTP POST requests for the Swift Object Server."""
        device, partition, account, container, obj, policy = \
            get_obj_name_and_placement(request)
        timing_stats_labels['account'] = account
        timing_stats_labels['container'] = container
        timing_stats_labels['policy'] = int(policy)

        req_timestamp = valid_timestamp(request)
        new_delete_at = int(request.headers.get('X-Delete-At') or 0)
        if new_delete_at and new_delete_at < req_timestamp:
            return HTTPBadRequest(body='X-Delete-At in past', request=request,
                                  content_type='text/plain')
        next_part_power = request.headers.get('X-Backend-Next-Part-Power')
        try:
            disk_file = self.get_diskfile(
                device, partition, account, container, obj,
                policy=policy, open_expired=is_backend_open_expired(request),
                next_part_power=next_part_power)
        except DiskFileDeviceUnavailable:
            return HTTPInsufficientStorage(drive=device, request=request)
        try:
            orig_metadata = disk_file.read_metadata(current_time=req_timestamp)
        except DiskFileXattrNotSupported:
            return HTTPInsufficientStorage(drive=device, request=request)
        except (DiskFileNotExist, DiskFileQuarantined):
            return HTTPNotFound(request=request)
        except DiskFileStateChanged:
            return HTTPServiceUnavailable(request=request)
        orig_timestamp = Timestamp(
            orig_metadata.get('X-Timestamp', Timestamp.zero()))
        orig_ctype_timestamp = disk_file.content_type_timestamp
        req_ctype_time = '0'
        req_ctype = request.headers.get('Content-Type')
        if req_ctype:
            req_ctype_time = request.headers.get('Content-Type-Timestamp',
                                                 req_timestamp.internal)
        req_ctype_timestamp = Timestamp(req_ctype_time)
        if orig_timestamp >= req_timestamp \
                and orig_ctype_timestamp >= req_ctype_timestamp:
            return HTTPConflict(
                request=request,
                headers={'X-Backend-Timestamp': orig_timestamp.internal})

        if req_timestamp > orig_timestamp:
            metadata = {'X-Timestamp': req_timestamp.internal}
            metadata.update(val for val in request.headers.items()
                            if (is_user_meta('object', val[0]) or
                                is_object_transient_sysmeta(val[0])))
            headers_to_copy = (
                request.headers.get(
                    'X-Backend-Replication-Headers', '').split() +
                list(self.allowed_headers))
            for header_key in headers_to_copy:
                if header_key in request.headers:
                    header_caps = bytes_to_wsgi(
                        wsgi_to_bytes(header_key).title())
                    metadata[header_caps] = request.headers[header_key]
            orig_delete_at = int(orig_metadata.get('X-Delete-At') or 0)
            disk_file_metadata = disk_file.get_datafile_metadata()
            self._conditional_delete_at_update(
                request, device, account, container, obj, policy,
                disk_file_metadata, orig_delete_at, new_delete_at
            )
        else:
            # preserve existing metadata, only content-type may be updated
            metadata = dict(disk_file.get_metafile_metadata())

        if req_ctype_timestamp > orig_ctype_timestamp:
            # we have a new content-type, add to metadata and container update
            content_type_headers = {
                'Content-Type': request.headers['Content-Type'],
                'Content-Type-Timestamp': req_ctype_timestamp.internal
            }
            metadata.update(content_type_headers)
        else:
            # send existing content-type with container update
            content_type_headers = {
                'Content-Type': disk_file.content_type,
                'Content-Type-Timestamp': orig_ctype_timestamp.internal
            }
            if orig_ctype_timestamp != disk_file.data_timestamp:
                # only add to metadata if it's not the datafile content-type
                metadata.update(content_type_headers)

        try:
            disk_file.write_metadata(metadata)
        except (DiskFileXattrNotSupported, DiskFileNoSpace):
            return HTTPInsufficientStorage(drive=device, request=request)

        if (content_type_headers['Content-Type-Timestamp']
                != disk_file.data_timestamp):
            # Current content-type is not from the datafile, but the datafile
            # content-type may have a swift_bytes param that was appended by
            # SLO and we must continue to send that with the container update.
            # Do this (rather than use a separate header) for backwards
            # compatibility because there may be 'legacy' container updates in
            # async pending that have content-types with swift_bytes params, so
            # we have to be able to handle those in container server anyway.
            _, swift_bytes = extract_swift_bytes(
                disk_file.get_datafile_metadata()['Content-Type'])
            if swift_bytes:
                content_type_headers['Content-Type'] += (';swift_bytes=%s'
                                                         % swift_bytes)

        update_headers = HeaderKeyDict({
            'x-size': orig_metadata['Content-Length'],
            'x-content-type': content_type_headers['Content-Type'],
            'x-timestamp': disk_file.data_timestamp.internal,
            'x-content-type-timestamp':
            content_type_headers['Content-Type-Timestamp'],
            'x-meta-timestamp': metadata['X-Timestamp'],
            'x-etag': orig_metadata['ETag']})

        # Special cases for backwards compatibility.
        # For EC policy, send X-Object-Sysmeta-Ec-Etag which is same as the
        # X-Backend-Container-Update-Override-Etag value sent with the original
        # PUT. Similarly send X-Object-Sysmeta-Ec-Content-Length which is the
        # same as the X-Backend-Container-Update-Override-Size value. We have
        # to send Etag and size with a POST container update because the
        # original PUT container update may have failed or be in async_pending.
        if 'X-Object-Sysmeta-Ec-Etag' in orig_metadata:
            update_headers['X-Etag'] = orig_metadata[
                'X-Object-Sysmeta-Ec-Etag']
        if 'X-Object-Sysmeta-Ec-Content-Length' in orig_metadata:
            update_headers['X-Size'] = orig_metadata[
                'X-Object-Sysmeta-Ec-Content-Length']

        self._check_container_override(update_headers, orig_metadata)

        # object POST updates are PUT to the container server
        self.container_update(
            'PUT', account, container, obj, request, update_headers,
            device, policy)

        # Add current content-type and sysmeta to response
        resp_headers = {
            'X-Backend-Content-Type': content_type_headers['Content-Type']}
        for key, value in orig_metadata.items():
            if is_sys_meta('object', key):
                resp_headers[key] = value

        return HTTPAccepted(request=request, headers=resp_headers)

    def _pre_create_checks(self, request, device, partition,
                           account, container, obj, policy):
        req_timestamp = valid_timestamp(request)
        error_response = check_object_creation(request, obj)
        if error_response:
            raise error_response
        try:
            fsize = request.message_length()
        except ValueError as e:
            raise HTTPBadRequest(body=str(e), request=request,
                                 content_type='text/plain')
        # In case of multipart-MIME put, the proxy sends a chunked request,
        # but may let us know the real content length so we can verify that
        # we have enough disk space to hold the object.
        if fsize is None:
            fsize = request.headers.get('X-Backend-Obj-Content-Length')
            if fsize is not None:
                try:
                    fsize = int(fsize)
                except ValueError as e:
                    raise HTTPBadRequest(body=str(e), request=request,
                                         content_type='text/plain')
        # SSYNC will include Frag-Index header for subrequests, in which case
        # get_diskfile will ignore non-matching on-disk data files
        frag_index = request.headers.get('X-Backend-Ssync-Frag-Index')
        next_part_power = request.headers.get('X-Backend-Next-Part-Power')
        try:
            disk_file = self.get_diskfile(
                device, partition, account, container, obj,
                policy=policy, frag_index=frag_index,
                next_part_power=next_part_power)
        except DiskFileDeviceUnavailable:
            raise HTTPInsufficientStorage(drive=device, request=request)
        try:
            orig_metadata = disk_file.read_metadata(current_time=req_timestamp)
            orig_timestamp = disk_file.data_timestamp
        except DiskFileXattrNotSupported:
            raise HTTPInsufficientStorage(drive=device, request=request)
        except DiskFileDeleted as e:
            orig_metadata = {}
            orig_timestamp = e.timestamp
        except (DiskFileNotExist, DiskFileQuarantined,
                DiskFileStateChanged):
            orig_metadata = {}
            orig_timestamp = Timestamp.zero()
        # Checks for If-None-Match
        if request.if_none_match is not None and orig_metadata:
            if '*' in request.if_none_match:
                # File exists already so return 412
                raise HTTPPreconditionFailed(request=request)
            if orig_metadata.get('ETag') in request.if_none_match:
                # The current ETag matches, so raise 412
                raise HTTPPreconditionFailed(request=request)
        if orig_timestamp >= req_timestamp:
            raise HTTPConflict(
                request=request,
                headers={'X-Backend-Timestamp': orig_timestamp.internal})
        return disk_file, fsize, orig_metadata

    def _do_multi_stage_mime_continue_headers(self, request, obj_input):
        """
        If the proxy wants to send us object metadata after the object body, it
        sets some headers. We have to tell the proxy, in the 100 Continue
        response, that we're able to parse a multipart MIME document and
        extract the object and metadata from it. If we don't, then the proxy
        won't actually send the footer metadata.

        If the proxy doesn't want to do any of that, this is the identity
        function for obj_input and multi_stage_mime_state will be False-y.

        :returns: a tuple, (obj_input, multi_stage_mime_state)
        """
        have_metadata_footer = False
        use_multiphase_commit = False

        hundred_continue_headers = []
        if config_true_value(
                request.headers.get(
                    'X-Backend-Obj-Multiphase-Commit')):
            use_multiphase_commit = True
            hundred_continue_headers.append(
                ('X-Obj-Multiphase-Commit', 'yes'))

        if config_true_value(
                request.headers.get('X-Backend-Obj-Metadata-Footer')):
            have_metadata_footer = True
            hundred_continue_headers.append(
                ('X-Obj-Metadata-Footer', 'yes'))

        if have_metadata_footer or use_multiphase_commit:
            obj_input.set_hundred_continue_response_headers(
                hundred_continue_headers)
            mime_boundary = wsgi_to_bytes(request.headers.get(
                'X-Backend-Obj-Multipart-Mime-Boundary'))
            if not mime_boundary:
                raise HTTPBadRequest("no MIME boundary")

            with ChunkReadTimeout(self.client_timeout):
                mime_documents_iter = iter_mime_headers_and_bodies(
                    request.environ['wsgi.input'],
                    mime_boundary, self.network_chunk_size)
                _junk_hdrs, obj_input = next(mime_documents_iter)
            multi_stage_mime_state = {
                'have_metadata_footer': have_metadata_footer,
                'use_multiphase_commit': use_multiphase_commit,
                'mime_documents_iter': mime_documents_iter,
            }
        else:
            multi_stage_mime_state = {}
        return obj_input, multi_stage_mime_state

    def _stage_obj_data(self, request, device, obj_input, writer, fsize):
        """
        Feed the object_input into the writer.

        :returns: a tuple, (upload_size, etag)
        """
        writer.open()
        elapsed_time = 0
        upload_expiration = time.time() + self.max_upload_time
        timeout_reader = self._make_timeout_reader(obj_input)

        # Wrap the chunks in CooperativeIterator with specified period
        cooperative_reader = CooperativeIterator(
            iter(timeout_reader, b''), period=self.cooperative_period
        )

        for chunk in cooperative_reader:
            start_time = time.time()
            if start_time > upload_expiration:
                self.logger.increment('PUT.timeouts')
                raise HTTPRequestTimeout(request=request)
            writer.write(chunk)
            elapsed_time += time.time() - start_time

        upload_size, etag = writer.chunks_finished()
        if fsize is not None and fsize != upload_size:
            raise HTTPClientDisconnect(request=request)
        if upload_size:
            self.logger.transfer_rate(
                'PUT.' + device + '.timing', elapsed_time,
                upload_size)
        return upload_size, etag

    def _get_request_metadata(self, request, upload_size, etag):
        """
        Pull object metadata off the request.

        :returns: metadata, a dict of object metadata
        """
        metadata = {
            'X-Timestamp': request.timestamp.internal,
            'Content-Type': request.headers['content-type'],
            'Content-Length': str(upload_size),
            'ETag': etag,
        }
        metadata.update(val for val in request.headers.items()
                        if (is_sys_or_user_meta('object', val[0]) or
                            is_object_transient_sysmeta(val[0])))
        headers_to_copy = (
            request.headers.get(
                'X-Backend-Replication-Headers', '').split() +
            list(self.allowed_headers))
        for header_key in headers_to_copy:
            if header_key in request.headers:
                header_caps = bytes_to_wsgi(
                    wsgi_to_bytes(header_key).title())
                metadata[header_caps] = request.headers[header_key]
        return metadata

    def _read_mime_footers_metadata(self, have_metadata_footer,
                                    mime_documents_iter, **kwargs):
        """
        Read footer metadata from the bottom of the multi-stage MIME body.

        :returns: metadata, a dict
        """
        if have_metadata_footer:
            metadata = self._read_metadata_footer(
                mime_documents_iter)
            footer_etag = metadata.pop('etag', '').lower()
            if footer_etag:
                metadata['ETag'] = footer_etag
        else:
            metadata = {}
        return metadata

    def _apply_extra_metadata(self, request, metadata, footers_metadata):
        """
        Apply extra metadata precedence to prepare metadata for storage.
        """
        metadata.update(val for val in footers_metadata.items()
                        if (is_sys_or_user_meta('object', val[0]) or
                            is_object_transient_sysmeta(val[0])))
        # N.B. footers_metadata is a HeaderKeyDict
        received_etag = normalize_etag(footers_metadata.get(
            'etag', request.headers.get('etag', '')))
        if received_etag and received_etag != metadata['ETag']:
            raise HTTPUnprocessableEntity(request=request)

    def _send_multi_stage_continue_headers(self, request,
                                           use_multiphase_commit,
                                           mime_documents_iter, **kwargs):
        """
        If the PUT requires a two-phase commit (a data and a commit phase) send
        the proxy server another 100-continue response to indicate that we are
        finished writing object data
        """
        if use_multiphase_commit:
            request.environ['wsgi.input'].\
                send_hundred_continue_response()
            if not self._read_put_commit_message(mime_documents_iter):
                raise HTTPServerError(request=request)

    def _drain_mime_request(self, mime_documents_iter, **kwargs):
        """
        Drain any remaining MIME docs from the socket. There shouldn't be any,
        but we must read the whole request body.
        """
        try:
            while True:
                with ChunkReadTimeout(self.client_timeout):
                    _junk_hdrs, _junk_body = next(mime_documents_iter)
                drain(_junk_body, self.network_chunk_size,
                      self.client_timeout)
        except ChunkReadError:
            raise HTTPClientDisconnect()
        except ChunkReadTimeout:
            raise HTTPRequestTimeout()
        except StopIteration:
            pass

    def _post_commit_updates(self, request, device,
                             account, container, obj, policy,
                             orig_metadata, footers_metadata, metadata):
        orig_delete_at = int(orig_metadata.get('X-Delete-At') or 0)
        new_delete_at = int(request.headers.get('X-Delete-At') or 0)

        self._conditional_delete_at_update(request, device, account, container,
                                           obj, policy, metadata,
                                           orig_delete_at, new_delete_at)

        update_headers = HeaderKeyDict({
            'x-size': metadata['Content-Length'],
            'x-content-type': metadata['Content-Type'],
            'x-timestamp': metadata['X-Timestamp'],
            'x-etag': metadata['ETag']})
        # apply any container update header overrides sent with request
        self._check_container_override(update_headers, request.headers,
                                       footers_metadata)
        self.container_update(
            'PUT', account, container, obj, request,
            update_headers, device, policy)

    @public
    @timing_stats()
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def PUT(self, request, timing_stats_labels):
        """Handle HTTP PUT requests for the Swift Object Server."""
        device, partition, account, container, obj, policy = \
            get_obj_name_and_placement(request)
        timing_stats_labels['account'] = account
        timing_stats_labels['container'] = container
        timing_stats_labels['policy'] = int(policy)

        disk_file, fsize, orig_metadata = self._pre_create_checks(
            request, device, partition, account, container, obj, policy)
        writer = disk_file.writer(size=fsize)
        try:
            obj_input = request.environ['wsgi.input']
            obj_input, multi_stage_mime_state = \
                self._do_multi_stage_mime_continue_headers(request, obj_input)
            upload_size, etag = self._stage_obj_data(
                request, device, obj_input, writer, fsize)
            metadata = self._get_request_metadata(request, upload_size, etag)
            if multi_stage_mime_state:
                footers_metadata = self._read_mime_footers_metadata(
                    **multi_stage_mime_state)
            else:
                footers_metadata = {}
            self._apply_extra_metadata(request, metadata, footers_metadata)
            writer.put(metadata)
            if multi_stage_mime_state:
                self._send_multi_stage_continue_headers(
                    request, **multi_stage_mime_state)
            if not config_true_value(
                    request.headers.get('X-Backend-No-Commit', False)):
                writer.commit(request.timestamp)
            if multi_stage_mime_state:
                self._drain_mime_request(**multi_stage_mime_state)
        except (DiskFileXattrNotSupported, DiskFileNoSpace):
            return HTTPInsufficientStorage(drive=device, request=request)
        except ChunkReadError:
            return HTTPClientDisconnect(request=request)
        except ChunkReadTimeout:
            return HTTPRequestTimeout(request=request)
        finally:
            writer.close()
        self._post_commit_updates(request, device,
                                  account, container, obj, policy,
                                  orig_metadata, footers_metadata, metadata)
        return HTTPCreated(request=request, etag=etag)

    @public
    @timing_stats()
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def GET(self, request, timing_stats_labels):
        """Handle HTTP GET requests for the Swift Object Server."""
        device, partition, account, container, obj, policy = \
            get_obj_name_and_placement(request)
        timing_stats_labels['account'] = account
        timing_stats_labels['container'] = container
        timing_stats_labels['policy'] = int(policy)

        request.headers.setdefault('X-Timestamp',
                                   normalize_timestamp(time.time()))
        req_timestamp = valid_timestamp(request)
        frag_prefs = safe_json_loads(
            request.headers.get('X-Backend-Fragment-Preferences'))
        try:
            disk_file = self.get_diskfile(
                device, partition, account, container, obj,
                policy=policy, frag_prefs=frag_prefs,
                open_expired=is_backend_open_expired(request))
        except DiskFileDeviceUnavailable:
            return HTTPInsufficientStorage(drive=device, request=request)
        try:
            with disk_file.open(current_time=req_timestamp):
                metadata = disk_file.get_metadata()
                resolve_ignore_range_header(request, metadata)
                obj_size = int(metadata['Content-Length'])
                file_x_ts = Timestamp(metadata['X-Timestamp'])
                keep_cache = (
                    self.keep_cache_private
                    or (
                        "X-Auth-Token" not in request.headers
                        and "X-Storage-Token" not in request.headers
                    )
                    or (
                        self.keep_cache_slo_manifest
                        and config_true_value(
                            metadata.get("X-Static-Large-Object")
                        )
                    )
                )
                conditional_etag = resolve_etag_is_at_header(request, metadata)
                app_iter = disk_file.reader(
                    keep_cache=keep_cache,
                    cooperative_period=self.cooperative_period,
                    etag_validate_frac=self.etag_validate_frac,
                )
                response = Response(
                    app_iter=app_iter, request=request,
                    conditional_response=True,
                    conditional_etag=conditional_etag,
                )
                response.headers['Content-Type'] = metadata.get(
                    'Content-Type', 'application/octet-stream')
                for key, value in metadata.items():
                    if (is_sys_or_user_meta('object', key) or
                            is_object_transient_sysmeta(key) or
                            key.lower() in self.allowed_headers):
                        response.headers[key] = value
                response.etag = metadata['ETag']
                response.last_modified = file_x_ts
                response.content_length = obj_size
                try:
                    response.content_encoding = metadata[
                        'Content-Encoding']
                except KeyError:
                    pass
                response.headers['X-Timestamp'] = file_x_ts.normal
                response.headers['X-Backend-Timestamp'] = file_x_ts.internal
                response.headers['X-Backend-Data-Timestamp'] = \
                    disk_file.data_timestamp.internal
                if disk_file.durable_timestamp:
                    response.headers['X-Backend-Durable-Timestamp'] = \
                        disk_file.durable_timestamp.internal
                response.headers['X-Backend-Fragments'] = \
                    _make_backend_fragments_header(disk_file.fragments)
                resp = request.get_response(response)
        except DiskFileXattrNotSupported:
            return HTTPInsufficientStorage(drive=device, request=request)
        except (DiskFileNotExist, DiskFileQuarantined) as e:
            headers = {}
            if hasattr(e, 'timestamp'):
                headers['X-Backend-Timestamp'] = e.timestamp.internal
            resp = HTTPNotFound(request=request, headers=headers,
                                conditional_response=True)
        except DiskFileStateChanged:
            resp = HTTPServiceUnavailable(request=request)
        return resp

    @public
    @timing_stats(sample_rate=0.8)
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def HEAD(self, request, timing_stats_labels):
        """Handle HTTP HEAD requests for the Swift Object Server."""
        device, partition, account, container, obj, policy = \
            get_obj_name_and_placement(request)
        timing_stats_labels['account'] = account
        timing_stats_labels['container'] = container
        timing_stats_labels['policy'] = int(policy)

        request.headers.setdefault('X-Timestamp',
                                   normalize_timestamp(time.time()))
        req_timestamp = valid_timestamp(request)
        frag_prefs = safe_json_loads(
            request.headers.get('X-Backend-Fragment-Preferences'))
        try:
            disk_file = self.get_diskfile(
                device, partition, account, container, obj,
                policy=policy, frag_prefs=frag_prefs,
                open_expired=is_backend_open_expired(request))
        except DiskFileDeviceUnavailable:
            return HTTPInsufficientStorage(drive=device, request=request)
        try:
            metadata = disk_file.read_metadata(current_time=req_timestamp)
        except DiskFileXattrNotSupported:
            return HTTPInsufficientStorage(drive=device, request=request)
        except (DiskFileNotExist, DiskFileQuarantined) as e:
            headers = {}
            if hasattr(e, 'timestamp'):
                headers['X-Backend-Timestamp'] = e.timestamp.internal
            return HTTPNotFound(request=request, headers=headers,
                                conditional_response=True)
        except DiskFileStateChanged:
            return HTTPServiceUnavailable(request=request)
        conditional_etag = resolve_etag_is_at_header(request, metadata)
        response = Response(request=request, conditional_response=True,
                            conditional_etag=conditional_etag)
        response.headers['Content-Type'] = metadata.get(
            'Content-Type', 'application/octet-stream')
        for key, value in metadata.items():
            if (is_sys_or_user_meta('object', key) or
                    is_object_transient_sysmeta(key) or
                    key.lower() in self.allowed_headers):
                response.headers[key] = value
        response.etag = metadata['ETag']
        ts = Timestamp(metadata['X-Timestamp'])
        response.last_modified = ts
        # Needed for container sync feature
        response.headers['X-Timestamp'] = ts.normal
        response.headers['X-Backend-Timestamp'] = ts.internal
        response.headers['X-Backend-Data-Timestamp'] = \
            disk_file.data_timestamp.internal
        if disk_file.durable_timestamp:
            response.headers['X-Backend-Durable-Timestamp'] = \
                disk_file.durable_timestamp.internal
        response.headers['X-Backend-Fragments'] = \
            _make_backend_fragments_header(disk_file.fragments)
        response.content_length = int(metadata['Content-Length'])
        try:
            response.content_encoding = metadata['Content-Encoding']
        except KeyError:
            pass
        return response

    @public
    @timing_stats()
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def DELETE(self, request, timing_stats_labels):
        """Handle HTTP DELETE requests for the Swift Object Server."""
        device, partition, account, container, obj, policy = \
            get_obj_name_and_placement(request)
        timing_stats_labels['account'] = account
        timing_stats_labels['container'] = container
        timing_stats_labels['policy'] = int(policy)

        req_timestamp = valid_timestamp(request)
        next_part_power = request.headers.get('X-Backend-Next-Part-Power')
        try:
            disk_file = self.get_diskfile(
                device, partition, account, container, obj,
                policy=policy, next_part_power=next_part_power)
        except DiskFileDeviceUnavailable:
            return HTTPInsufficientStorage(drive=device, request=request)
        try:
            orig_metadata = disk_file.read_metadata(current_time=req_timestamp)
        except DiskFileXattrNotSupported:
            return HTTPInsufficientStorage(drive=device, request=request)
        except DiskFileExpired as e:
            orig_timestamp = e.timestamp
            orig_metadata = e.metadata
            response_class = HTTPNotFound
        except DiskFileDeleted as e:
            orig_timestamp = e.timestamp
            orig_metadata = {}
            response_class = HTTPNotFound
        except (DiskFileNotExist, DiskFileQuarantined):
            orig_timestamp = 0
            orig_metadata = {}
            response_class = HTTPNotFound
        except DiskFileStateChanged:
            # Bail out early because unread metadata may have an X-Delete-At.
            # This condition can occur when a concurrent request has replaced
            # the data or meta file that this thread was trying to open. Force
            # client to retry, hoping the on-disk files have settled.
            return HTTPServiceUnavailable(request=request)
        else:
            orig_timestamp = disk_file.data_timestamp
            if orig_timestamp < req_timestamp:
                response_class = HTTPNoContent
            else:
                response_class = HTTPConflict
        response_timestamp = max(orig_timestamp, req_timestamp)
        orig_delete_at = Timestamp(
            orig_metadata.get('X-Delete-At') or Timestamp.zero())
        try:
            req_if_delete_at_val = request.headers['x-if-delete-at']
            req_if_delete_at = Timestamp(req_if_delete_at_val)
        except KeyError:
            pass
        except ValueError:
            return HTTPBadRequest(
                request=request,
                body='Bad X-If-Delete-At header value')
        else:
            # request includes x-if-delete-at; we must not place a tombstone
            # if we can not verify the x-if-delete-at time
            if not orig_timestamp:
                # no object found at all
                return HTTPNotFound()
            if orig_timestamp >= req_timestamp:
                # Found a newer object -- return 409 as work item is stale
                return HTTPConflict()
            if orig_delete_at != req_if_delete_at:
                return HTTPPreconditionFailed(
                    request=request,
                    body='X-If-Delete-At and X-Delete-At do not match')
            else:
                # differentiate success from no object at all
                response_class = HTTPNoContent
        self._conditional_delete_at_update(
            request, device, account, container, obj, policy, {},
            orig_delete_at, 0
        )
        if orig_timestamp < req_timestamp:
            try:
                disk_file.delete(req_timestamp)
            except DiskFileNoSpace:
                return HTTPInsufficientStorage(drive=device, request=request)
            self.container_update(
                'DELETE', account, container, obj, request,
                HeaderKeyDict({'x-timestamp': req_timestamp.internal}),
                device, policy)
        return response_class(
            request=request,
            headers={'X-Backend-Timestamp': response_timestamp.internal,
                     'X-Backend-Content-Type': orig_metadata.get(
                         'Content-Type', '')})

    @public
    @replication
    @timing_stats(sample_rate=0.1)
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def REPLICATE(self, request, timing_stats_labels):
        """
        Handle REPLICATE requests for the Swift Object Server.  This is used
        by the object replicator to get hashes for directories.

        Note that the name REPLICATE is preserved for historical reasons as
        this verb really just returns the hashes information for the specified
        parameters and is used, for example, by both replication and EC.
        """
        device, partition, suffix_parts, policy = \
            get_name_and_placement(request, 2, 3, True)
        suffixes = suffix_parts.split('-') if suffix_parts else []
        skip_rehash = bool(suffixes)
        timing_stats_labels['policy'] = int(policy)
        timing_stats_labels['skip_rehash'] = skip_rehash

        try:
            hashes = self._diskfile_router[policy].get_hashes(
                device, partition, suffixes, policy,
                skip_rehash=skip_rehash)
        except DiskFileDeviceUnavailable:
            resp = HTTPInsufficientStorage(drive=device, request=request)
        else:
            # force pickle protocol for compatibility with py2 nodes
            resp = Response(body=pickle.dumps(hashes, protocol=2))
        return resp

    @public
    @replication
    @timing_stats(sample_rate=0.1)
    @labeled_timing_stats(metric=LABELED_METRIC_NAME)
    def SSYNC(self, request, timing_stats_labels):
        # the ssync sender may want to send PUT subrequests for non-durable
        # data that should not be committed; legacy behaviour has been to
        # commit all PUTs (subject to EC footer metadata), so we need to
        # indicate to the sender that this object server has been upgraded to
        # understand the X-Backend-No-Commit header.
        headers = {'X-Backend-Accept-No-Commit': True}
        receiver = ssync_receiver.Receiver(self, request)
        timing_stats_labels['policy'] = int(receiver.policy)
        return Response(app_iter=receiver(), headers=headers)

    def __call__(self, env, start_response):
        """WSGI Application entry point for the Swift Object Server."""
        start_time = time.time()
        req = Request(env)
        self.logger.txn_id = req.headers.get('x-trans-id', None)

        if not check_utf8(wsgi_to_str(req.path_info), internal=True):
            res = HTTPPreconditionFailed(body='Invalid UTF8 or contains NULL')
        else:
            try:
                # disallow methods which have not been marked 'public'
                if req.method not in self.allowed_methods:
                    res = HTTPMethodNotAllowed()
                else:
                    res = getattr(self, req.method)(req)
            except DiskFileCollision:
                res = HTTPForbidden(request=req)
            except HTTPException as error_response:
                res = error_response
            except (Exception, Timeout):
                self.logger.exception(
                    'ERROR __call__ error with %(method)s'
                    ' %(path)s ', {'method': req.method, 'path': req.path})
                res = HTTPInternalServerError(body=traceback.format_exc())
        trans_time = time.time() - start_time
        res.fix_conditional_response()
        if self.log_requests:
            log_line = get_log_line(req, res, trans_time, '', self.log_format,
                                    self.anonymization_method,
                                    self.anonymization_salt)
            if req.method in ('REPLICATE', 'SSYNC') or \
                    'X-Backend-Replication' in req.headers:
                self.logger.debug(log_line)
            else:
                self.logger.info(log_line)
        if req.method in ('PUT', 'DELETE'):
            slow = self.slow - trans_time
            if slow > 0:
                sleep(slow)

        # To be able to zero-copy send the object, we need a few things.
        # First, we have to be responding successfully to a GET, or else we're
        # not sending the object. Second, we have to be able to extract the
        # socket file descriptor from the WSGI input object. Third, the
        # diskfile has to support zero-copy send.
        #
        # There's a good chance that this could work for 206 responses too,
        # but the common case is sending the whole object, so we'll start
        # there.
        if req.method == 'GET' and res.status_int == 200 and \
           isinstance(env['wsgi.input'], wsgi.Input):
            app_iter = getattr(res, 'app_iter', None)
            checker = getattr(app_iter, 'can_zero_copy_send', None)
            if checker and checker():
                # For any kind of zero-copy thing like sendfile or splice, we
                # need the file descriptor. Eventlet doesn't provide a clean
                # way of getting that, so we resort to this.
                wsock = env['wsgi.input'].get_socket()
                wsockfd = wsock.fileno()

                # Don't call zero_copy_send() until after we force the HTTP
                # headers out of Eventlet and into the socket.
                def zero_copy_iter():
                    # If possible, set TCP_CORK so that headers don't
                    # immediately go on the wire, but instead, wait for some
                    # response body to make the TCP frames as large as
                    # possible (and hence as few packets as possible).
                    #
                    # On non-Linux systems, we might consider TCP_NODELAY, but
                    # since the only known zero-copy-capable diskfile uses
                    # Linux-specific syscalls, we'll defer that work until
                    # someone needs it.
                    if hasattr(socket, 'TCP_CORK'):
                        wsock.setsockopt(socket.IPPROTO_TCP,
                                         socket.TCP_CORK, 1)
                    yield EventletPlungerString()
                    try:
                        app_iter.zero_copy_send(wsockfd)
                    except Exception:
                        self.logger.exception("zero_copy_send() blew up")
                        raise
                    yield b''

                # Get headers ready to go out
                res(env, start_response)
                return zero_copy_iter()
            else:
                return res(env, start_response)
        else:
            return res(env, start_response)


def global_conf_callback(preloaded_app_conf, global_conf):
    """
    Callback for swift.common.wsgi.run_wsgi during the global_conf
    creation so that we can add our replication_semaphore, used to
    limit the number of concurrent SSYNC_REQUESTS across all
    workers.

    :param preloaded_app_conf: The preloaded conf for the WSGI app.
                               This conf instance will go away, so
                               just read from it, don't write.
    :param global_conf: The global conf that will eventually be
                        passed to the app_factory function later.
                        This conf is created before the worker
                        subprocesses are forked, so can be useful to
                        set up semaphores, shared memory, etc.
    """
    replication_concurrency = int(
        preloaded_app_conf.get('replication_concurrency') or 4)
    if replication_concurrency:
        # Have to put the value in a list so it can get past paste
        global_conf['replication_semaphore'] = [
            multiprocessing.BoundedSemaphore(replication_concurrency)]


def app_factory(global_conf, **local_conf):
    """paste.deploy app factory for creating WSGI object server apps"""
    conf = global_conf.copy()
    conf.update(local_conf)
    return ObjectController(conf)


def main():
    conf_file, options = parse_options(test_config=True)
    sys.exit(run_wsgi(conf_file, 'object-server',
                      global_conf_callback=global_conf_callback,
                      **options))


if __name__ == '__main__':
    main()
